"""Test help"""


def test_missing_command(runner, yadm_y):
    """Run without any command"""
    run = runner(command=yadm_y())
    assert run.failure
    assert run.err == ''
    assert run.out.startswith('Usage: yadm')


def test_help_command(runner, yadm_y):
    """Run with help command"""
    run = runner(command=yadm_y('help'))
    assert run.failure
    assert run.err == ''
    assert run.out.startswith('Usage: yadm')



"""Unit tests: record_template"""

INIT_VARS = """
    alt_targets=()
    alt_template_cmds=()
    alt_sources=()
"""

REPORT_RESULTS = """
    echo "SIZE:${#alt_targets[@]}"
    echo "TARGETS:${alt_targets[@]}"
    echo "CMDS:${alt_template_cmds[@]}"
    echo "SOURCES:${alt_sources[@]}"
"""


def test_new_template(runner, yadm):
    """Test new template"""

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        record_template "tgt_one"   "cmd_one"   "src_one"
        record_template "tgt_two"   "cmd_two"   "src_two"
        record_template "tgt_three" "cmd_three" "src_three"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:3\n' in run.out
    assert 'TARGETS:tgt_one tgt_two tgt_three\n' in run.out
    assert 'CMDS:cmd_one cmd_two cmd_three\n' in run.out
    assert 'SOURCES:src_one src_two src_three\n' in run.out


def test_existing_template(runner, yadm):
    """Overwrite existing templates"""

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        alt_targets=("testtgt")
        alt_template_cmds=("existing_cmd")
        alt_sources=("existing_src")
        record_template "testtgt" "new_cmd" "new_src"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:1\n' in run.out
    assert 'TARGETS:testtgt\n' in run.out
    assert 'CMDS:new_cmd\n' in run.out
    assert 'SOURCES:new_src\n' in run.out



"""Test encryption"""

import os
import pipes
import time
import pytest

KEY_FILE = 'test/test_key'
KEY_FINGERPRINT = 'F8BBFC746C58945442349BCEBA54FFD04C599B1A'
KEY_NAME = 'yadm-test1'
KEY_TRUST = 'test/ownertrust.txt'
PASSPHRASE = 'ExamplePassword'

pytestmark = pytest.mark.usefixtures('config_git')


def add_asymmetric_key(runner, gnupg):
    """Add asymmetric key"""
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    runner(
        ['gpg', '--import', pipes.quote(KEY_FILE)],
        env=env,
        shell=True,
    )
    runner(
        ['gpg', '--import-ownertrust', '<', pipes.quote(KEY_TRUST)],
        env=env,
        shell=True,
    )


def remove_asymmetric_key(runner, gnupg):
    """Remove asymmetric key"""
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    runner(
        [
            'gpg', '--batch', '--yes',
            '--delete-secret-keys', pipes.quote(KEY_FINGERPRINT)
        ],
        env=env,
        shell=True,
    )
    runner(
        [
            'gpg', '--batch', '--yes',
            '--delete-key', pipes.quote(KEY_FINGERPRINT)
        ],
        env=env,
        shell=True,
    )


@pytest.fixture
def asymmetric_key(runner, gnupg):
    """Fixture for asymmetric key, removed in teardown"""
    add_asymmetric_key(runner, gnupg)
    yield KEY_NAME
    remove_asymmetric_key(runner, gnupg)


@pytest.fixture
def encrypt_targets(yadm_y, paths):
    """Fixture for setting up data to encrypt

    This fixture:
      * inits an empty repo
      * creates test files in the work tree
      * creates a ".yadm/encrypt" file for testing:
        * standard files
        * standard globs
        * directories
        * comments
        * empty lines and lines with just space
        * exclusions
      * returns a list of expected encrypted files
    """

    # init empty yadm repo
    os.system(' '.join(yadm_y('init', '-w', str(paths.work), '-f')))

    expected = []

    # standard files w/ dirs & spaces
    paths.work.join('inc file1').write('inc file1')
    expected.append('inc file1')
    paths.encrypt.write('inc file1\n')
    paths.work.join('inc dir').mkdir()
    paths.work.join('inc dir/inc file2').write('inc file2')
    expected.append('inc dir/inc file2')
    paths.encrypt.write('inc dir/inc file2\n', mode='a')

    # standard globs w/ dirs & spaces
    paths.work.join('globs file1').write('globs file1')
    expected.append('globs file1')
    paths.work.join('globs dir').mkdir()
    paths.work.join('globs dir/globs file2').write('globs file2')
    expected.append('globs dir/globs file2')
    paths.encrypt.write('globs*\n', mode='a')

    # blank lines
    paths.encrypt.write('\n        \n\t\n', mode='a')

    # comments
    paths.work.join('commentfile1').write('commentfile1')
    paths.encrypt.write('#commentfile1\n', mode='a')
    paths.encrypt.write('        #commentfile1\n', mode='a')

    # exclusions
    paths.work.join('extest').mkdir()
    paths.encrypt.write('extest/*\n', mode='a')  # include within extest
    paths.work.join('extest/inglob1').write('inglob1')
    paths.work.join('extest/exglob1').write('exglob1')
    paths.work.join('extest/exglob2').write('exglob2')
    paths.encrypt.write('!extest/ex*\n', mode='a')  # exclude the ex*
    expected.append('extest/inglob1')  # should be left with only in*

    return expected


@pytest.fixture(scope='session')
def decrypt_targets(tmpdir_factory, runner, gnupg):
    """Fixture for setting data to decrypt

    This fixture:
      * creates symmetric/asymmetric encrypted archives
      * creates a list of expected decrypted files
    """

    tmpdir = tmpdir_factory.mktemp('decrypt_targets')
    symmetric = tmpdir.join('symmetric.tar.gz.gpg')
    asymmetric = tmpdir.join('asymmetric.tar.gz.gpg')

    expected = []

    tmpdir.join('decrypt1').write('decrypt1')
    expected.append('decrypt1')
    tmpdir.join('decrypt2').write('decrypt2')
    expected.append('decrypt2')
    tmpdir.join('subdir').mkdir()
    tmpdir.join('subdir/decrypt3').write('subdir/decrypt3')
    expected.append('subdir/decrypt3')

    gnupg.pw(PASSPHRASE)
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    run = runner(
        ['tar', 'cvf', '-'] +
        expected +
        ['|', 'gpg', '--batch', '--yes', '-c'] +
        ['--output', pipes.quote(str(symmetric))],
        cwd=tmpdir,
        env=env,
        shell=True)
    assert run.success

    gnupg.pw('')
    add_asymmetric_key(runner, gnupg)
    run = runner(
        ['tar', 'cvf', '-'] +
        expected +
        ['|', 'gpg', '--batch', '--yes', '-e'] +
        ['-r', pipes.quote(KEY_NAME)] +
        ['--output', pipes.quote(str(asymmetric))],
        cwd=tmpdir,
        env=env,
        shell=True)
    assert run.success
    remove_asymmetric_key(runner, gnupg)

    return {
        'asymmetric': asymmetric,
        'expected': expected,
        'symmetric': symmetric,
    }


@pytest.mark.parametrize(
    'bad_phrase', [False, True],
    ids=['good_phrase', 'bad_phrase'])
@pytest.mark.parametrize(
    'missing_encrypt', [False, True],
    ids=['encrypt_exists', 'encrypt_missing'])
@pytest.mark.parametrize(
    'overwrite', [False, True],
    ids=['clean', 'overwrite'])
def test_symmetric_encrypt(
        runner, yadm_y, paths, encrypt_targets,
        gnupg, bad_phrase, overwrite, missing_encrypt):
    """Test symmetric encryption"""

    if missing_encrypt:
        paths.encrypt.remove()

    if bad_phrase:
        gnupg.pw('')
    else:
        gnupg.pw(PASSPHRASE)

    if overwrite:
        paths.archive.write('existing archive')

    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    run = runner(yadm_y('encrypt'), env=env)

    if missing_encrypt or bad_phrase:
        assert run.failure
    else:
        assert run.success
        assert run.err == ''

    if missing_encrypt:
        assert 'does not exist' in run.out
    elif bad_phrase:
        assert 'Invalid passphrase' in run.err
    else:
        assert encrypted_data_valid(
            runner, gnupg, paths.archive, encrypt_targets)


@pytest.mark.parametrize(
    'bad_phrase', [False, True],
    ids=['good_phrase', 'bad_phrase'])
@pytest.mark.parametrize(
    'archive_exists', [True, False],
    ids=['archive_exists', 'archive_missing'])
@pytest.mark.parametrize(
    'dolist', [False, True],
    ids=['decrypt', 'list'])
def test_symmetric_decrypt(
        runner, yadm_y, paths, decrypt_targets, gnupg,
        dolist, archive_exists, bad_phrase):
    """Test decryption"""

    # init empty yadm repo
    os.system(' '.join(yadm_y('init', '-w', str(paths.work), '-f')))

    if bad_phrase:
        gnupg.pw('')
        time.sleep(1)  # allow gpg-agent cache to expire
    else:
        gnupg.pw(PASSPHRASE)

    if archive_exists:
        decrypt_targets['symmetric'].copy(paths.archive)

    # to test overwriting
    paths.work.join('decrypt1').write('pre-existing file')

    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home

    args = []

    if dolist:
        args.append('-l')
    run = runner(yadm_y('decrypt') + args, env=env)

    if archive_exists and not bad_phrase:
        assert run.success
        assert 'encrypted with 1 passphrase' in run.err
        if dolist:
            for filename in decrypt_targets['expected']:
                if filename != 'decrypt1':  # this one should exist
                    assert not paths.work.join(filename).exists()
                assert filename in run.out
        else:
            for filename in decrypt_targets['expected']:
                assert paths.work.join(filename).read() == filename
    else:
        assert run.failure


@pytest.mark.usefixtures('asymmetric_key')
@pytest.mark.parametrize(
    'ask', [False, True],
    ids=['no_ask', 'ask'])
@pytest.mark.parametrize(
    'key_exists', [True, False],
    ids=['key_exists', 'key_missing'])
@pytest.mark.parametrize(
    'overwrite', [False, True],
    ids=['clean', 'overwrite'])
def test_asymmetric_encrypt(
        runner, yadm_y, paths, encrypt_targets, gnupg,
        overwrite, key_exists, ask):
    """Test asymmetric encryption"""

    # specify encryption recipient
    if ask:
        os.system(' '.join(yadm_y('config', 'yadm.gpg-recipient', 'ASK')))
        expect = [('Enter the user ID', KEY_NAME), ('Enter the user ID', '')]
    else:
        os.system(' '.join(yadm_y('config', 'yadm.gpg-recipient', KEY_NAME)))
        expect = []

    if overwrite:
        paths.archive.write('existing archive')

    if not key_exists:
        remove_asymmetric_key(runner, gnupg)

    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home

    run = runner(yadm_y('encrypt'), env=env, expect=expect)

    if key_exists:
        assert run.success
        assert encrypted_data_valid(
            runner, gnupg, paths.archive, encrypt_targets)
    else:
        assert run.failure
        assert 'Unable to write' in run.out

    if ask:
        assert 'Enter the user ID' in run.out


@pytest.mark.usefixtures('asymmetric_key')
@pytest.mark.parametrize(
    'key_exists', [True, False],
    ids=['key_exists', 'key_missing'])
@pytest.mark.parametrize(
    'dolist', [False, True],
    ids=['decrypt', 'list'])
def test_asymmetric_decrypt(
        runner, yadm_y, paths, decrypt_targets, gnupg,
        dolist, key_exists):
    """Test decryption"""

    # init empty yadm repo
    os.system(' '.join(yadm_y('init', '-w', str(paths.work), '-f')))

    decrypt_targets['asymmetric'].copy(paths.archive)

    # to test overwriting
    paths.work.join('decrypt1').write('pre-existing file')

    if not key_exists:
        remove_asymmetric_key(runner, gnupg)

    args = []

    if dolist:
        args.append('-l')
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    run = runner(yadm_y('decrypt') + args, env=env)

    if key_exists:
        assert run.success
        if dolist:
            for filename in decrypt_targets['expected']:
                if filename != 'decrypt1':  # this one should exist
                    assert not paths.work.join(filename).exists()
                assert filename in run.out
        else:
            for filename in decrypt_targets['expected']:
                assert paths.work.join(filename).read() == filename
    else:
        assert run.failure
        assert 'Unable to extract encrypted files' in run.out


@pytest.mark.parametrize(
    'untracked',
    [False, 'y', 'n'],
    ids=['tracked', 'untracked_answer_y', 'untracked_answer_n'])
def test_offer_to_add(
        runner, yadm_y, paths, encrypt_targets, gnupg, untracked):
    """Test offer to add encrypted archive

    All the other encryption tests use an archive outside of the work tree.
    However, the archive is often inside the work tree, and if it is, there
    should be an offer to add it to the repo if it is not tracked.
    """

    worktree_archive = paths.work.join('worktree-archive.tar.gpg')

    expect = []

    gnupg.pw(PASSPHRASE)
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home

    if untracked:
        expect.append(('add it now', untracked))
    else:
        worktree_archive.write('exists')
        os.system(' '.join(yadm_y('add', str(worktree_archive))))

    run = runner(
        yadm_y('encrypt', '--yadm-archive', str(worktree_archive)),
        env=env,
        expect=expect
        )

    assert run.success
    assert run.err == ''
    assert encrypted_data_valid(
        runner, gnupg, worktree_archive, encrypt_targets)

    run = runner(
        yadm_y('status', '--porcelain', '-uall', str(worktree_archive)))
    assert run.success
    assert run.err == ''

    if untracked == 'y':
        # should be added to the index
        assert f'A  {worktree_archive.basename}' in run.out
    elif untracked == 'n':
        # should NOT be added to the index
        assert f'?? {worktree_archive.basename}' in run.out
    else:
        # should appear modified in the index
        assert f'AM {worktree_archive.basename}' in run.out


@pytest.mark.usefixtures('ds1_copy')
def test_encrypt_added_to_exclude(runner, yadm_y, paths, gnupg):
    """Confirm that .config/yadm/encrypt is added to exclude"""

    gnupg.pw(PASSPHRASE)
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home

    exclude_file = paths.repo.join('info/exclude')
    paths.encrypt.write('test-encrypt-data\n')
    paths.work.join('test-encrypt-data').write('')
    exclude_file.write('original-data', ensure=True)

    run = runner(yadm_y('encrypt'), env=env)

    assert 'test-encrypt-data' in paths.repo.join('info/exclude').read()
    assert 'original-data' in paths.repo.join('info/exclude').read()
    assert run.success
    assert run.err == ''


def encrypted_data_valid(runner, gnupg, encrypted, expected):
    """Verify encrypted data matches expectations"""
    gnupg.pw(PASSPHRASE)
    env = os.environ.copy()
    env['GNUPGHOME'] = gnupg.home
    run = runner([
        'gpg',
        '-d', pipes.quote(str(encrypted)),
        '2>/dev/null',
        '|', 'tar', 't'], env=env, shell=True, report=False)
    file_count = 0
    for filename in run.out.splitlines():
        if filename.endswith('/'):
            continue
        file_count += 1
        assert filename in expected, (
            f'Unexpected file in archive: {filename}')
    assert file_count == len(expected), (
        'Number of files in archive does not match expected')
    return True



"""Unit tests: record_score"""
import pytest

INIT_VARS = """
    score=0
    local_class=testclass
    local_system=testsystem
    local_host=testhost
    local_user=testuser
    alt_scores=()
    alt_targets=()
    alt_sources=()
    alt_template_cmds=()
"""

REPORT_RESULTS = """
    echo "SIZE:${#alt_scores[@]}"
    echo "SCORES:${alt_scores[@]}"
    echo "TARGETS:${alt_targets[@]}"
    echo "SOURCES:${alt_sources[@]}"
"""


def test_dont_record_zeros(runner, yadm):
    """Record nothing if the score is zero"""

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        record_score "0" "testtgt" "testsrc"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:0\n' in run.out
    assert 'SCORES:\n' in run.out
    assert 'TARGETS:\n' in run.out
    assert 'SOURCES:\n' in run.out


def test_new_scores(runner, yadm):
    """Test new scores"""

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        record_score "1" "tgt_one"   "src_one"
        record_score "2" "tgt_two"   "src_two"
        record_score "4" "tgt_three" "src_three"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:3\n' in run.out
    assert 'SCORES:1 2 4\n' in run.out
    assert 'TARGETS:tgt_one tgt_two tgt_three\n' in run.out
    assert 'SOURCES:src_one src_two src_three\n' in run.out


@pytest.mark.parametrize('difference', ['lower', 'equal', 'higher'])
def test_existing_scores(runner, yadm, difference):
    """Test existing scores"""

    expected_score = '2'
    expected_src = 'existing_src'
    if difference == 'lower':
        score = '1'
    elif difference == 'equal':
        score = '2'
    else:
        score = '4'
        expected_score = '4'
        expected_src = 'new_src'

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        alt_scores=(2)
        alt_targets=("testtgt")
        alt_sources=("existing_src")
        record_score "{score}" "testtgt" "new_src"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:1\n' in run.out
    assert f'SCORES:{expected_score}\n' in run.out
    assert 'TARGETS:testtgt\n' in run.out
    assert f'SOURCES:{expected_src}\n' in run.out


def test_existing_template(runner, yadm):
    """Record nothing if a template command is registered for this target"""

    script = f"""
        YADM_TEST=1 source {yadm}
        {INIT_VARS}
        alt_scores=(1)
        alt_targets=("testtgt")
        alt_sources=()
        alt_template_cmds=("existing_template")
        record_score "2" "testtgt" "new_src"
        {REPORT_RESULTS}
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert 'SIZE:1\n' in run.out
    assert 'SCORES:1\n' in run.out
    assert 'TARGETS:testtgt\n' in run.out
    assert 'SOURCES:\n' in run.out



"""Unit tests: upgrade"""
import pytest

LEGACY_PATHS = [
    'config',
    'encrypt',
    'files.gpg',
    'bootstrap',
    'hooks/pre_command',
    'hooks/post_command',
]

# used:
# YADM_COMPATIBILITY
# YADM_DIR
# YADM_LEGACY_DIR
# GIT_PROGRAM
@pytest.mark.parametrize('condition', ['compat', 'equal', 'existing_repo'])
def test_upgrade_errors(tmpdir, runner, yadm, condition):
    """Test upgrade() error conditions"""

    compatibility = 'YADM_COMPATIBILITY=1' if condition == 'compat' else ''

    home = tmpdir.mkdir('home')
    yadm_dir = home.join('.config/yadm')
    legacy_dir = home.join('.yadm')
    if condition == 'equal':
        legacy_dir = yadm_dir
    if condition == 'existing_repo':
        yadm_dir.ensure_dir('repo.git')
        legacy_dir.ensure_dir('repo.git')

    script = f"""
        YADM_TEST=1 source {yadm}
        {compatibility}
        YADM_DIR="{yadm_dir}"
        YADM_REPO="{yadm_dir}/repo.git"
        YADM_LEGACY_DIR="{legacy_dir}"
        upgrade
    """
    run = runner(command=['bash'], inp=script)
    assert run.failure
    assert run.err == ''
    assert 'Unable to upgrade' in run.out
    if condition == 'compat':
        assert 'YADM_COMPATIBILITY' in run.out
    if condition == 'equal':
        assert 'has been resolved as' in run.out
    if condition == 'existing_repo':
        assert 'already exists' in run.out


@pytest.mark.parametrize(
    'condition', ['no-paths', 'untracked', 'tracked', 'submodules'])
def test_upgrade(tmpdir, runner, yadm, condition):
    """Test upgrade()

    When testing the condition of git-tracked data, "echo" will be used as a
    mock for git. echo will return true, simulating a positive result from "git
    ls-files". Also echo will report the parameters for "git mv".
    """
    home = tmpdir.mkdir('home')
    yadm_dir = home.join('.config/yadm')
    legacy_dir = home.join('.yadm')

    if condition != 'no-paths':
        legacy_dir.join('repo.git/config').write('test-repo', ensure=True)
        for lpath in LEGACY_PATHS:
            legacy_dir.join(lpath).write(lpath, ensure=True)

    mock_git = ""
    if condition in ['tracked', 'submodules']:
        mock_git = f'''
            function git() {{
                echo "$@"
                if [[ "$*" == *.gitmodules* ]]; then
                    return { '0' if condition == 'submodules' else '1' }
                fi
                return 0
            }}
        '''

    script = f"""
        YADM_TEST=1 source {yadm}
        YADM_DIR="{yadm_dir}"
        YADM_REPO="{yadm_dir}/repo.git"
        YADM_LEGACY_DIR="{legacy_dir}"
        GIT_PROGRAM="git"
        {mock_git}
        function cd {{ echo "$@";}}
        upgrade
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if condition == 'no-paths':
        assert 'Upgrade is not necessary' in run.out
    else:
        for lpath in LEGACY_PATHS + ['repo.git']:
            expected = (
                f'Moving {legacy_dir.join(lpath)} '
                f'to {yadm_dir.join(lpath)}')
            assert expected in run.out
        if condition == 'untracked':
            assert 'test-repo' in yadm_dir.join('repo.git/config').read()
            for lpath in LEGACY_PATHS:
                assert lpath in yadm_dir.join(lpath).read()
        elif condition in ['tracked', 'submodules']:
            for lpath in LEGACY_PATHS:
                expected = (
                    f'mv {legacy_dir.join(lpath)} '
                    f'{yadm_dir.join(lpath)}')
                assert expected in run.out
            assert 'files tracked by yadm have been renamed' in run.out
            if condition == 'submodules':
                assert 'submodule deinit -f .' in run.out
                assert 'submodule update --init --recursive' in run.out
            else:
                assert 'submodule deinit -f .' not in run.out
                assert 'submodule update --init --recursive' not in run.out



"""Test version"""

import re
import pytest


@pytest.fixture(scope='module')
def expected_version(yadm):
    """
    Expected semantic version number. This is taken directly out of yadm,
    searching for the VERSION= string.
    """
    yadm_version = re.findall(
        r'VERSION=([^\n]+)',
        open(yadm).read())
    if yadm_version:
        return yadm_version[0]
    pytest.fail(f'version not found in {yadm}')
    return 'not found'


def test_semantic_version(expected_version):
    """Version is semantic"""
    # semantic version conforms to MAJOR.MINOR.PATCH
    assert re.search(r'^\d+\.\d+\.\d+$', expected_version), (
        'does not conform to MAJOR.MINOR.PATCH')


def test_reported_version(
        runner, yadm_y, expected_version):
    """Report correct version"""
    run = runner(command=yadm_y('version'))
    assert run.success
    assert run.err == ''
    assert run.out == f'yadm {expected_version}\n'



"""Testing Utilities

This module holds values/functions common to multiple tests.
"""

import re
import os

ALT_FILE1 = 'test_alt'
ALT_FILE2 = 'test alt/test alt'
ALT_DIR = 'test alt/test alt dir'

# Directory based alternates must have a tracked contained file.
# This will be the test contained file name
CONTAINED = 'contained_file'

# These variables are used for making include files which will be processed
# within jinja templates
INCLUDE_FILE = 'inc_file'
INCLUDE_DIRS = ['', 'test alt']
INCLUDE_CONTENT = '8780846c02e34c930d0afd127906668f'


def set_local(paths, variable, value):
    """Set local override"""
    os.system(
        f'GIT_DIR={str(paths.repo)} '
        f'git config --local "local.{variable}" "{value}"'
    )


def create_alt_files(paths, suffix,
                     preserve=False, tracked=True,
                     encrypt=False, exclude=False,
                     content=None, includefile=False,
                     yadm_alt=False, yadm_dir=None):
    """Create new files, and add to the repo

    This is used for testing alternate files. In each case, a suffix is
    appended to two standard file paths. Particulars of the file creation and
    repo handling are dependent upon the function arguments.
    """

    basepath = yadm_dir.join('alt') if yadm_alt else paths.work

    if not preserve:
        for remove_path in (ALT_FILE1, ALT_FILE2, ALT_DIR):
            if basepath.join(remove_path).exists():
                basepath.join(remove_path).remove(rec=1, ignore_errors=True)
                assert not basepath.join(remove_path).exists()

    new_file1 = basepath.join(ALT_FILE1 + suffix)
    new_file1.write(ALT_FILE1 + suffix, ensure=True)
    new_file2 = basepath.join(ALT_FILE2 + suffix)
    new_file2.write(ALT_FILE2 + suffix, ensure=True)
    new_dir = basepath.join(ALT_DIR + suffix).join(CONTAINED)
    new_dir.write(ALT_DIR + suffix, ensure=True)

    # Do not test directory support for jinja alternates
    test_paths = [new_file1, new_file2]
    test_names = [ALT_FILE1, ALT_FILE2]
    if not re.match(r'##(t$|t\.|template|yadm)', suffix):
        test_paths += [new_dir]
        test_names += [ALT_DIR]

    for test_path in test_paths:
        if content:
            test_path.write('\n' + content, mode='a', ensure=True)
        assert test_path.exists()

    _create_includefiles(includefile, test_paths, basepath)
    _create_tracked(tracked, test_paths, paths)

    prefix = '.config/yadm/alt/' if yadm_alt else ''
    _create_encrypt(encrypt, test_names, suffix, paths, exclude, prefix)


def parse_alt_output(output, linked=True):
    """Parse output of 'alt', and return list of linked files"""
    regex = r'Creating (.+) from template (.+)$'
    if linked:
        regex = r'Linking (.+) to (.+)$'
    parsed_list = dict()
    for line in output.splitlines():
        match = re.match(regex, line)
        if match:
            if linked:
                parsed_list[match.group(2)] = match.group(1)
            else:
                parsed_list[match.group(1)] = match.group(2)
    return parsed_list.values()


def _create_includefiles(includefile, test_paths, basepath):
    if includefile:
        for dpath in INCLUDE_DIRS:
            incfile = basepath.join(dpath + '/' + INCLUDE_FILE)
            incfile.write(INCLUDE_CONTENT, ensure=True)
            test_paths += [incfile]


def _create_tracked(tracked, test_paths, paths):
    if tracked:
        for track_path in test_paths:
            os.system(f'GIT_DIR={str(paths.repo)} git add "{track_path}"')
        os.system(f'GIT_DIR={str(paths.repo)} git commit -m "Add test files"')


def _create_encrypt(encrypt, test_names, suffix, paths, exclude, prefix):
    if encrypt:
        for encrypt_name in test_names:
            paths.encrypt.write(
                f'{prefix + encrypt_name + suffix}\n', mode='a')
            if exclude:
                paths.encrypt.write(
                    f'!{prefix + encrypt_name + suffix}\n', mode='a')



"""Test enter"""

import os
import pytest


@pytest.mark.parametrize(
    'shell, success', [
        ('delete', True),  # if there is no shell variable, bash creates it
        ('', False),
        ('/usr/bin/env', True),
        ('noexec', False),
    ], ids=[
        'shell-missing',
        'shell-empty',
        'shell-env',
        'shell-noexec',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_enter(runner, yadm_y, paths, shell, success):
    """Enter tests"""
    env = os.environ.copy()
    if shell == 'delete':
        # remove shell
        if 'SHELL' in env:
            del env['SHELL']
    elif shell == 'noexec':
        # specify a non-executable path
        noexec = paths.root.join('noexec')
        noexec.write('')
        noexec.chmod(0o664)
        env['SHELL'] = str(noexec)
    else:
        env['SHELL'] = shell

    run = runner(command=yadm_y('enter'), env=env)
    assert run.success == success
    assert run.err == ''
    prompt = f'yadm shell ({paths.repo})'
    if success:
        assert run.out.startswith('Entering yadm repo')
        assert run.out.rstrip().endswith('Leaving yadm repo')
    if not success:
        assert 'does not refer to an executable' in run.out
    if 'env' in shell:
        assert f'GIT_DIR={paths.repo}' in run.out
        assert f'GIT_WORK_TREE={paths.work}' in run.out
        assert f'PROMPT={prompt}' in run.out
        assert f'PS1={prompt}' in run.out


@pytest.mark.parametrize(
    'shell, opts, path', [
        ('bash', '--norc', '\\w'),
        ('csh', '-f', '%~'),
        ('zsh', '-f', '%~'),
    ], ids=[
        'bash',
        'csh',
        'zsh',
    ])
@pytest.mark.parametrize(
    'cmd',
    [False, 'cmd', 'cmd-bad-exit'],
    ids=['no-cmd', 'cmd', 'cmd-bad-exit'])
@pytest.mark.usefixtures('ds1_copy')
def test_enter_shell_ops(runner, yadm_y, paths, shell, opts, path, cmd):
    """Enter tests for specific shell options"""

    change_exit = '\nfalse' if cmd == 'cmd-bad-exit' else ''

    # Create custom shell to detect options passed
    custom_shell = paths.root.join(shell)
    custom_shell.write(
        f'#!/bin/sh\necho OPTS=$*\necho PROMPT=$PROMPT{change_exit}'
    )
    custom_shell.chmod(0o775)

    test_cmd = ['test1', 'test2', 'test3']

    enter_cmd = ['enter']
    if cmd:
        enter_cmd += test_cmd

    env = os.environ.copy()
    env['SHELL'] = custom_shell

    run = runner(command=yadm_y(*enter_cmd), env=env)
    if cmd == 'cmd-bad-exit':
        assert run.failure
    else:
        assert run.success
    assert run.err == ''
    assert f'OPTS={opts}' in run.out
    assert f'PROMPT=yadm shell ({paths.repo}) {path} >' in run.out
    if cmd:
        assert '-c ' + ' '.join(test_cmd) in run.out
        assert 'Entering yadm repo' not in run.out
        assert 'Leaving yadm repo' not in run.out
    else:
        assert 'Entering yadm repo' in run.out
        assert 'Leaving yadm repo' in run.out



"""Test jinja"""

import os
import pytest
import utils

# These tests are for the template processing in YADM_COMPATIBILITY=1 mode
pytestmark = pytest.mark.deprecated


@pytest.fixture(scope='module')
def envtpl_present(runner):
    """Is envtpl present and working?"""
    try:
        run = runner(command=['envtpl', '-h'])
        if run.success:
            return True
    except OSError:
        pass
    return False


@pytest.mark.usefixtures('ds1_copy')
def test_local_override(runner, yadm_y, paths,
                        tst_distro, envtpl_present):
    """Test local overrides"""
    if not envtpl_present:
        pytest.skip('Unable to test without envtpl.')

    # define local overrides
    utils.set_local(paths, 'class', 'or-class')
    utils.set_local(paths, 'hostname', 'or-hostname')
    utils.set_local(paths, 'os', 'or-os')
    utils.set_local(paths, 'user', 'or-user')

    template = (
        'j2-{{ YADM_CLASS }}-'
        '{{ YADM_OS }}-{{ YADM_HOSTNAME }}-'
        '{{ YADM_USER }}-{{ YADM_DISTRO }}'
        '-{%- '
        f"include '{utils.INCLUDE_FILE}'"
        ' -%}'
    )
    expected = (
        f'j2-or-class-or-os-or-hostname-or-user-{tst_distro}'
        f'-{utils.INCLUDE_CONTENT}'
    )

    utils.create_alt_files(paths, '##yadm.j2', content=template,
                           includefile=True)

    # os.system(f'find {paths.work}' + ' -name *j2 -ls -exec cat \'{}\' ";"')
    # os.system(f'find {paths.work}')
    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    created = utils.parse_alt_output(run.out, linked=False)

    # assert the proper creation has occurred
    for file_path in (utils.ALT_FILE1, utils.ALT_FILE2):
        source_file = file_path + '##yadm.j2'
        assert paths.work.join(file_path).isfile()
        lines = paths.work.join(file_path).readlines(cr=False)
        assert lines[0] == source_file
        assert lines[1] == expected
        assert str(paths.work.join(source_file)) in created


@pytest.mark.parametrize('autoalt', [None, 'true', 'false'])
@pytest.mark.usefixtures('ds1_copy')
def test_auto_alt(runner, yadm_y, paths, autoalt, tst_sys,
                  envtpl_present):
    """Test setting auto-alt"""

    if not envtpl_present:
        pytest.skip('Unable to test without envtpl.')

    # set the value of auto-alt
    if autoalt:
        os.system(' '.join(yadm_y('config', 'yadm.auto-alt', autoalt)))

    # create file
    jinja_suffix = '##yadm.j2'
    utils.create_alt_files(paths, jinja_suffix, content='{{ YADM_OS }}')

    # run status to possibly trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('status'), env=env)
    assert run.success
    assert run.err == ''
    created = utils.parse_alt_output(run.out, linked=False)

    # assert the proper creation has occurred
    for file_path in (utils.ALT_FILE1, utils.ALT_FILE2):
        source_file = file_path + jinja_suffix
        if autoalt == 'false':
            assert not paths.work.join(file_path).exists()
        else:
            assert paths.work.join(file_path).isfile()
            lines = paths.work.join(file_path).readlines(cr=False)
            assert lines[0] == source_file
            assert lines[1] == tst_sys
            # no created output when run via auto-alt
            assert str(paths.work.join(source_file)) not in created


@pytest.mark.usefixtures('ds1_copy')
def test_jinja_envtpl_missing(runner, paths):
    """Test operation when envtpl is missing"""

    script = f"""
        YADM_TEST=1 source {paths.pgm}
        process_global_args -Y "{paths.yadm}"
        set_operating_system
        configure_paths
        YADM_COMPATIBILITY=1
        ENVTPL_PROGRAM='envtpl_missing' main alt
    """

    utils.create_alt_files(paths, '##yadm.j2')

    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert f'envtpl not available, not creating' in run.out


@pytest.mark.parametrize(
    'tracked, encrypt, exclude', [
        (False, False, False),
        (True, False, False),
        (False, True, False),
        (False, True, True),
    ], ids=[
        'untracked',
        'tracked',
        'encrypted',
        'excluded',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_jinja(runner, yadm_y, paths,
               tst_sys, tst_host, tst_user, tst_distro,
               tracked, encrypt, exclude,
               envtpl_present):
    """Test jinja processing"""

    if not envtpl_present:
        pytest.skip('Unable to test without envtpl.')

    jinja_suffix = '##yadm.j2'

    # set the class
    tst_class = 'testclass'
    utils.set_local(paths, 'class', tst_class)

    template = (
        'j2-{{ YADM_CLASS }}-'
        '{{ YADM_OS }}-{{ YADM_HOSTNAME }}-'
        '{{ YADM_USER }}-{{ YADM_DISTRO }}'
        '-{%- '
        f"include '{utils.INCLUDE_FILE}'"
        ' -%}'
    )
    expected = (
        f'j2-{tst_class}-'
        f'{tst_sys}-{tst_host}-'
        f'{tst_user}-{tst_distro}'
        f'-{utils.INCLUDE_CONTENT}'
    )

    utils.create_alt_files(paths, jinja_suffix, content=template,
                           tracked=tracked, encrypt=encrypt, exclude=exclude,
                           includefile=True)

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    created = utils.parse_alt_output(run.out, linked=False)

    # assert the proper creation has occurred
    for file_path in (utils.ALT_FILE1, utils.ALT_FILE2):
        source_file = file_path + jinja_suffix
        if tracked or (encrypt and not exclude):
            assert paths.work.join(file_path).isfile()
            lines = paths.work.join(file_path).readlines(cr=False)
            assert lines[0] == source_file
            assert lines[1] == expected
            assert str(paths.work.join(source_file)) in created
        else:
            assert not paths.work.join(file_path).exists()
            assert str(paths.work.join(source_file)) not in created



"""Test git-crypt"""

import pytest


@pytest.mark.parametrize(
    'crypt',
    [False, 'installed', 'installed-but-failed'],
    ids=['not-installed', 'installed', 'installed-but-failed']
)
def test_git_crypt(runner, yadm, paths, tmpdir, crypt):
    """git-crypt tests"""

    paths.repo.ensure(dir=True)
    bindir = tmpdir.mkdir('bin')
    pgm = bindir.join('test-git-crypt')

    if crypt:
        pgm.write(f'#!/bin/sh\necho git-crypt ran\n')
        pgm.chmod(0o775)
    if crypt == 'installed-but-failed':
        pgm.write('false\n', mode='a')

    script = f"""
        YADM_TEST=1 source {yadm}
        YADM_REPO={paths.repo}
        GIT_CRYPT_PROGRAM="{pgm}"
        git_crypt "param1"
    """

    run = runner(command=['bash'], inp=script)

    if crypt:
        if crypt == 'installed-but-failed':
            assert run.failure
        else:
            assert run.success
        assert run.out.strip() == 'git-crypt ran'
    else:
        assert run.failure
        assert f"command '{pgm}' cannot be located" in run.out
    assert run.err == ''



"""Unit tests: set_local_alt_values"""
import pytest
import utils


@pytest.mark.parametrize(
    'override', [
        False,
        'class',
        'os',
        'hostname',
        'user',
        ],
    ids=[
        'no-override',
        'override-class',
        'override-os',
        'override-hostname',
        'override-user',
        ]
    )
@pytest.mark.usefixtures('ds1_copy')
def test_set_local_alt_values(
        runner, yadm, paths, tst_sys, tst_host, tst_user, override):
    """Use issue_legacy_path_warning"""
    script = f"""
        YADM_TEST=1 source {yadm} &&
        set_operating_system &&
        YADM_DIR={paths.yadm} configure_paths &&
        set_local_alt_values
        echo "class='$local_class'"
        echo "os='$local_system'"
        echo "host='$local_host'"
        echo "user='$local_user'"
    """

    if override:
        utils.set_local(paths, override, 'override')

    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''

    if override == 'class':
        assert "class='override'" in run.out
    else:
        assert "class=''" in run.out

    if override == 'os':
        assert "os='override'" in run.out
    else:
        assert f"os='{tst_sys}'" in run.out

    if override == 'hostname':
        assert f"host='override'" in run.out
    else:
        assert f"host='{tst_host}'" in run.out

    if override == 'user':
        assert f"user='override'" in run.out
    else:
        assert f"user='{tst_user}'" in run.out


def test_distro(runner, yadm):
    """Assert that local_distro is set"""

    script = f"""
        YADM_TEST=1 source {yadm}
        function query_distro() {{ echo "testdistro"; }}
        set_local_alt_values
        echo "distro='$local_distro'"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out.strip() == "distro='testdistro'"



"""Unit tests: exclude_encrypted"""
import pytest


@pytest.mark.parametrize(
    'exclude', ['missing', 'outdated', 'up-to-date'])
@pytest.mark.parametrize(
    'encrypt_exists', [True, False], ids=['encrypt', 'no-encrypt'])
@pytest.mark.parametrize(
    'auto_exclude', [True, False], ids=['enabled', 'disabled'])
def test_exclude_encrypted(
        runner, tmpdir, yadm, encrypt_exists, auto_exclude, exclude):
    """Test exclude_encrypted()"""

    header = (
        "# yadm-auto-excludes\n"
        "# This section is managed by yadm.\n"
        "# Any edits below will be lost.\n"
    )

    config_function = 'function config() { echo "false";}'
    if auto_exclude:
        config_function = 'function config() { return; }'

    encrypt_file = tmpdir.join('encrypt_file')
    repo_dir = tmpdir.join('repodir')
    exclude_file = repo_dir.join('info/exclude')

    if encrypt_exists:
        encrypt_file.write('test-encrypt-data\n', ensure=True)
    if exclude == 'outdated':
        exclude_file.write(
            f'original-exclude\n{header}outdated\n', ensure=True)
    elif exclude == 'up-to-date':
        exclude_file.write(
            f'original-exclude\n{header}test-encrypt-data\n', ensure=True)

    script = f"""
        YADM_TEST=1 source {yadm}
        {config_function}
        DEBUG=1
        YADM_ENCRYPT="{encrypt_file}"
        YADM_REPO="{repo_dir}"
        exclude_encrypted
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''

    if auto_exclude:
        if encrypt_exists:
            assert exclude_file.exists()
            if exclude == 'missing':
                assert exclude_file.read() == f'{header}test-encrypt-data\n'
            else:
                assert exclude_file.read() == (
                    'original-exclude\n'
                    f'{header}test-encrypt-data\n')
            if exclude != 'up-to-date':
                assert f'Updating {exclude_file}' in run.out
            else:
                assert run.out == ''
        else:
            assert run.out == ''
    else:
        assert run.out == ''



"""Test bootstrap"""

import pytest


@pytest.mark.parametrize(
    'exists, executable, code, expect', [
        (False, False, 1, 'Cannot execute bootstrap'),
        (True, False, 1, 'is not an executable program'),
        (True, True, 123, 'Bootstrap successful'),
    ], ids=[
        'missing',
        'not executable',
        'executable',
    ])
def test_bootstrap(
        runner, yadm_y, paths, exists, executable, code, expect):
    """Test bootstrap command"""
    if exists:
        paths.bootstrap.write('')
    if executable:
        paths.bootstrap.write(
            '#!/bin/bash\n'
            f'echo {expect}\n'
            f'exit {code}\n'
        )
        paths.bootstrap.chmod(0o775)
    run = runner(command=yadm_y('bootstrap'))
    assert run.code == code
    assert run.err == ''
    assert expect in run.out



"""Unit tests: remove_stale_links"""
import os
import pytest


@pytest.mark.parametrize('linked', [True, False])
@pytest.mark.parametrize('kind', ['file', 'symlink'])
def test_remove_stale_links(runner, yadm, tmpdir, kind, linked):
    """Test remove_stale_links()"""

    source_file = tmpdir.join('source_file')
    source_file.write('source file', ensure=True)
    link = tmpdir.join('link')

    if kind == 'file':
        link.write('link file', ensure=True)
    else:
        os.system(f'ln -s {source_file} {link}')

    alt_linked = ''
    if linked:
        alt_linked = source_file

    script = f"""
        YADM_TEST=1 source {yadm}
        possible_alts=({link})
        alt_linked=({alt_linked})
        function rm() {{ echo rm "$@"; }}
        remove_stale_links
    """

    run = runner(command=['bash'], inp=script)
    assert run.err == ''
    if kind == 'symlink' and not linked:
        assert f'rm -f {link}' in run.out
    else:
        assert run.out == ''



"""Test clone"""

import os
import re
import pytest

BOOTSTRAP_CODE = 123
BOOTSTRAP_MSG = 'Bootstrap successful'


@pytest.mark.usefixtures('remote')
@pytest.mark.parametrize(
    'good_remote, repo_exists, force, conflicts', [
        (False, False, False, False),
        (True, False, False, False),
        (True, True, False, False),
        (True, True, True, False),
        (True, False, False, True),
    ], ids=[
        'bad remote',
        'simple',
        'existing repo',
        '-f',
        'conflicts',
    ])
def test_clone(
        runner, paths, yadm_y, repo_config, ds1,
        good_remote, repo_exists, force, conflicts):
    """Test basic clone operation"""

    # clear out the work path
    paths.work.remove()
    paths.work.mkdir()

    # determine remote url
    remote_url = f'file://{paths.remote}'
    if not good_remote:
        remote_url = 'file://bad_remote'

    old_repo = None
    if repo_exists:
        # put a repo in the way
        paths.repo.mkdir()
        old_repo = paths.repo.join('old_repo')
        old_repo.write('old_repo')

    if conflicts:
        ds1.tracked[0].relative.write('conflict')
        assert ds1.tracked[0].relative.exists()

    # run the clone command
    args = ['clone', '-w', paths.work]
    if force:
        args += ['-f']
    args += [remote_url]
    run = runner(command=yadm_y(*args))

    if not good_remote:
        # clone should fail
        assert run.failure
        assert run.err != ''
        assert 'Unable to fetch origin' in run.out
        assert not paths.repo.exists()
    elif repo_exists and not force:
        # can't overwrite data
        assert run.failure
        assert run.err == ''
        assert 'Git repo already exists' in run.out
    else:
        # clone should succeed, and repo should be configured properly
        assert successful_clone(run, paths, repo_config)

        # ensure conflicts are handled properly
        if conflicts:
            assert 'NOTE' in run.out
            assert 'Merging origin/master failed' in run.out
            assert 'Conflicts preserved' in run.out

        # confirm correct Git origin
        run = runner(
            command=('git', 'remote', '-v', 'show'),
            env={'GIT_DIR': paths.repo})
        assert run.success
        assert run.err == ''
        assert f'origin\t{remote_url}' in run.out

        # ensure conflicts are really preserved
        if conflicts:
            # test to see if the work tree is actually "clean"
            run = runner(
                command=yadm_y('status', '-uno', '--porcelain'),
                cwd=paths.work)
            assert run.success
            assert run.err == ''
            assert run.out == '', 'worktree has unexpected changes'

            # test to see if the conflicts are stashed
            run = runner(command=yadm_y('stash', 'list'), cwd=paths.work)
            assert run.success
            assert run.err == ''
            assert 'Conflicts preserved' in run.out, 'conflicts not stashed'

            # verify content of the stashed conflicts
            run = runner(command=yadm_y('stash', 'show', '-p'), cwd=paths.work)
            assert run.success
            assert run.err == ''
            assert '\n+conflict' in run.out, 'conflicts not stashed'

    # another force-related assertion
    if old_repo:
        if force:
            assert not old_repo.exists()
        else:
            assert old_repo.exists()


@pytest.mark.usefixtures('remote')
@pytest.mark.parametrize(
    'bs_exists, bs_param, answer', [
        (False, '--bootstrap', None),
        (True, '--bootstrap', None),
        (True, '--no-bootstrap', None),
        (True, None, 'n'),
        (True, None, 'y'),
    ], ids=[
        'force, missing',
        'force, existing',
        'prevent',
        'existing, answer n',
        'existing, answer y',
    ])
def test_clone_bootstrap(
        runner, paths, yadm_y, repo_config, bs_exists, bs_param, answer):
    """Test bootstrap clone features"""

    # establish a bootstrap
    create_bootstrap(paths, bs_exists)

    # run the clone command
    args = ['clone', '-w', paths.work]
    if bs_param:
        args += [bs_param]
    args += [f'file://{paths.remote}']
    expect = []
    if answer:
        expect.append(('Would you like to execute it now', answer))
    run = runner(command=yadm_y(*args), expect=expect)

    if answer:
        assert 'Would you like to execute it now' in run.out

    expected_code = 0
    if bs_exists and bs_param != '--no-bootstrap':
        expected_code = BOOTSTRAP_CODE

    if answer == 'y':
        expected_code = BOOTSTRAP_CODE
        assert BOOTSTRAP_MSG in run.out
    elif answer == 'n':
        expected_code = 0
        assert BOOTSTRAP_MSG not in run.out

    assert successful_clone(run, paths, repo_config, expected_code)

    if not bs_exists:
        assert BOOTSTRAP_MSG not in run.out


def create_bootstrap(paths, exists):
    """Create bootstrap file for test"""
    if exists:
        paths.bootstrap.write(
            '#!/bin/sh\n'
            f'echo {BOOTSTRAP_MSG}\n'
            f'exit {BOOTSTRAP_CODE}\n')
        paths.bootstrap.chmod(0o775)
        assert paths.bootstrap.exists()
    else:
        assert not paths.bootstrap.exists()


@pytest.mark.usefixtures('remote')
@pytest.mark.parametrize(
    'private_type, in_repo, in_work', [
        ('ssh', False, True),
        ('gnupg', False, True),
        ('ssh', True, True),
        ('gnupg', True, True),
        ('ssh', True, False),
        ('gnupg', True, False),
    ], ids=[
        'open ssh, not tracked',
        'open gnupg, not tracked',
        'open ssh, tracked',
        'open gnupg, tracked',
        'missing ssh, tracked',
        'missing gnupg, tracked',
    ])
def test_clone_perms(
        runner, yadm_y, paths, repo_config,
        private_type, in_repo, in_work):
    """Test clone permission-related functions"""

    # update remote repo to include private data
    if in_repo:
        rpath = paths.work.mkdir(f'.{private_type}').join('related')
        rpath.write('related')
        os.system(f'GIT_DIR="{paths.remote}" git add {rpath}')
        os.system(f'GIT_DIR="{paths.remote}" git commit -m "{rpath}"')
        rpath.remove()

    # ensure local private data is insecure at the start
    if in_work:
        pdir = paths.work.join(f'.{private_type}')
        if not pdir.exists():
            pdir.mkdir()
        pfile = pdir.join('existing')
        pfile.write('existing')
        pdir.chmod(0o777)
        pfile.chmod(0o777)
    else:
        paths.work.remove()
        paths.work.mkdir()

    env = {'HOME': paths.work}
    run = runner(
        yadm_y('clone', '-d', '-w', paths.work, f'file://{paths.remote}'),
        env=env
    )

    assert successful_clone(run, paths, repo_config)
    if in_work:
        # private directories which already exist, should be left as they are,
        # which in this test is "insecure".
        assert re.search(
            f'initial private dir perms drwxrwxrwx.+.{private_type}',
            run.out)
        assert re.search(
            f'pre-merge private dir perms drwxrwxrwx.+.{private_type}',
            run.out)
        assert re.search(
            f'post-merge private dir perms drwxrwxrwx.+.{private_type}',
            run.out)
    else:
        # private directories which are created, should be done prior to
        # merging, and with secure permissions.
        assert 'initial private dir perms' not in run.out
        assert re.search(
            f'pre-merge private dir perms drwx------.+.{private_type}',
            run.out)
        assert re.search(
            f'post-merge private dir perms drwx------.+.{private_type}',
            run.out)

    # standard perms still apply afterwards unless disabled with auto.perms
    assert oct(
        paths.work.join(f'.{private_type}').stat().mode).endswith('00'), (
            f'.{private_type} has not been secured by auto.perms')


@pytest.mark.usefixtures('remote')
@pytest.mark.parametrize('branch', ['master', 'valid', 'invalid'])
def test_alternate_branch(runner, paths, yadm_y, repo_config, branch):
    """Test cloning a branch other than master"""

    # add a "valid" branch to the remote
    os.system(f'GIT_DIR="{paths.remote}" git checkout -b valid')
    os.system(
        f'GIT_DIR="{paths.remote}" git commit '
        f'--allow-empty -m "This branch is valid"')

    # clear out the work path
    paths.work.remove()
    paths.work.mkdir()

    remote_url = f'file://{paths.remote}'

    # run the clone command
    args = ['clone', '-w', paths.work]
    if branch != 'master':
        args += ['-b', branch]
    args += [remote_url]
    run = runner(command=yadm_y(*args))

    if branch == 'invalid':
        assert run.failure
        assert 'ERROR: Clone failed' in run.out
        assert f"'origin/{branch}' does not exist in {remote_url}" in run.out
    else:
        assert successful_clone(run, paths, repo_config)

        # confirm correct Git origin
        run = runner(
            command=('git', 'remote', '-v', 'show'),
            env={'GIT_DIR': paths.repo})
        assert run.success
        assert run.err == ''
        assert f'origin\t{remote_url}' in run.out
        run = runner(command=yadm_y('show'))
        if branch == 'valid':
            assert 'This branch is valid' in run.out
        else:
            assert 'Initial commit' in run.out


def successful_clone(run, paths, repo_config, expected_code=0):
    """Assert clone is successful"""
    assert run.code == expected_code
    assert 'Initialized' in run.out
    assert oct(paths.repo.stat().mode).endswith('00'), 'Repo is not secured'
    assert repo_config('core.bare') == 'false'
    assert repo_config('status.showUntrackedFiles') == 'no'
    assert repo_config('yadm.managed') == 'true'
    return True


@pytest.fixture()
def remote(paths, ds1_repo_copy):
    """Function scoped remote (based on ds1)"""
    # pylint: disable=unused-argument
    # This is ignored because
    # @pytest.mark.usefixtures('ds1_remote_copy')
    # cannot be applied to another fixture.
    paths.remote.remove()
    paths.repo.move(paths.remote)



"""Test clean"""


def test_clean_command(runner, yadm_y):
    """Run with clean command"""
    run = runner(command=yadm_y('clean'))
    # do nothing, this is a dangerous Git command when managing dot files
    # report the command as disabled and exit as a failure
    assert run.failure
    assert run.err == ''
    assert 'disabled' in run.out



"""Unit tests: report_invalid_alts"""
import pytest


@pytest.mark.parametrize(
    'condition', [
        'compat',
        'previous-message',
        'invalid-alts',
        'no-invalid-alts',
    ])
def test_report_invalid_alts(runner, yadm, condition):
    """Use report_invalid_alts"""

    compat = ''
    previous = ''
    alts = 'INVALID_ALT=()'
    if condition == 'compat':
        compat = 'YADM_COMPATIBILITY=1'
    if condition == 'previous-message':
        previous = 'LEGACY_WARNING_ISSUED=1'
    if condition == 'invalid-alts':
        alts = 'INVALID_ALT=("file##invalid")'

    script = f"""
        YADM_TEST=1 source {yadm}
        {compat}
        {previous}
        {alts}
        report_invalid_alts
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if condition == 'invalid-alts':
        assert 'WARNING' in run.out
        assert 'file##invalid' in run.out
    else:
        assert run.out == ''



"""Test introspect"""

import pytest


@pytest.mark.parametrize(
    'name', [
        '',
        'invalid',
        'commands',
        'configs',
        'repo',
        'switches',
    ])
def test_introspect_category(
        runner, yadm_y, paths, name,
        supported_commands, supported_configs, supported_switches):
    """Validate introspection category"""
    if name:
        run = runner(command=yadm_y('introspect', name))
    else:
        run = runner(command=yadm_y('introspect'))

    assert run.success
    assert run.err == ''

    expected = []
    if name == 'commands':
        expected = supported_commands
    elif name == 'configs':
        expected = supported_configs
    elif name == 'switches':
        expected = supported_switches

    # assert values
    if name in ('', 'invalid'):
        assert run.out == ''
    if name == 'repo':
        assert run.out.rstrip() == paths.repo

    # make sure every expected value is present
    for value in expected:
        assert value in run.out
    # make sure nothing extra is present
    if expected:
        assert len(run.out.split()) == len(expected)



"""Unit tests: score_file"""
import pytest

CONDITION = {
    'default': {
        'labels': ['default'],
        'modifier': 0,
        },
    'system': {
        'labels': ['o', 'os'],
        'modifier': 1,
        },
    'distro': {
        'labels': ['d', 'distro'],
        'modifier': 2,
        },
    'class': {
        'labels': ['c', 'class'],
        'modifier': 4,
        },
    'hostname': {
        'labels': ['h', 'hostname'],
        'modifier': 8,
        },
    'user': {
        'labels': ['u', 'user'],
        'modifier': 16,
        },
    }
TEMPLATE_LABELS = ['t', 'template', 'yadm']


def calculate_score(filename):
    """Calculate the expected score"""
    # pylint: disable=too-many-branches
    score = 0

    _, conditions = filename.split('##', 1)

    for condition in conditions.split(','):
        label = condition
        value = None
        if '.' in condition:
            label, value = condition.split('.', 1)
        if label in CONDITION['default']['labels']:
            score += 1000
        elif label in CONDITION['system']['labels']:
            if value == 'testsystem':
                score += 1000 + CONDITION['system']['modifier']
            else:
                score = 0
                break
        elif label in CONDITION['distro']['labels']:
            if value == 'testdistro':
                score += 1000 + CONDITION['distro']['modifier']
            else:
                score = 0
                break
        elif label in CONDITION['class']['labels']:
            if value == 'testclass':
                score += 1000 + CONDITION['class']['modifier']
            else:
                score = 0
                break
        elif label in CONDITION['hostname']['labels']:
            if value == 'testhost':
                score += 1000 + CONDITION['hostname']['modifier']
            else:
                score = 0
                break
        elif label in CONDITION['user']['labels']:
            if value == 'testuser':
                score += 1000 + CONDITION['user']['modifier']
            else:
                score = 0
                break
        elif label in TEMPLATE_LABELS:
            score = 0
            break
    return score


@pytest.mark.parametrize(
    'default', ['default', None], ids=['default', 'no-default'])
@pytest.mark.parametrize(
    'system', ['system', None], ids=['system', 'no-system'])
@pytest.mark.parametrize(
    'distro', ['distro', None], ids=['distro', 'no-distro'])
@pytest.mark.parametrize(
    'cla', ['class', None], ids=['class', 'no-class'])
@pytest.mark.parametrize(
    'host', ['hostname', None], ids=['hostname', 'no-host'])
@pytest.mark.parametrize(
    'user', ['user', None], ids=['user', 'no-user'])
def test_score_values(
        runner, yadm, default, system, distro, cla, host, user):
    """Test score results"""
    # pylint: disable=too-many-branches
    local_class = 'testclass'
    local_system = 'testsystem'
    local_distro = 'testdistro'
    local_host = 'testhost'
    local_user = 'testuser'
    filenames = {'filename##': 0}

    if default:
        for filename in list(filenames):
            for label in CONDITION[default]['labels']:
                newfile = filename
                if not newfile.endswith('##'):
                    newfile += ','
                newfile += label
                filenames[newfile] = calculate_score(newfile)
    if system:
        for filename in list(filenames):
            for match in [True, False]:
                for label in CONDITION[system]['labels']:
                    newfile = filename
                    if not newfile.endswith('##'):
                        newfile += ','
                    newfile += '.'.join([
                        label,
                        local_system if match else 'badsys'
                        ])
                    filenames[newfile] = calculate_score(newfile)
    if distro:
        for filename in list(filenames):
            for match in [True, False]:
                for label in CONDITION[distro]['labels']:
                    newfile = filename
                    if not newfile.endswith('##'):
                        newfile += ','
                    newfile += '.'.join([
                        label,
                        local_distro if match else 'baddistro'
                        ])
                    filenames[newfile] = calculate_score(newfile)
    if cla:
        for filename in list(filenames):
            for match in [True, False]:
                for label in CONDITION[cla]['labels']:
                    newfile = filename
                    if not newfile.endswith('##'):
                        newfile += ','
                    newfile += '.'.join([
                        label,
                        local_class if match else 'badclass'
                        ])
                    filenames[newfile] = calculate_score(newfile)
    if host:
        for filename in list(filenames):
            for match in [True, False]:
                for label in CONDITION[host]['labels']:
                    newfile = filename
                    if not newfile.endswith('##'):
                        newfile += ','
                    newfile += '.'.join([
                        label,
                        local_host if match else 'badhost'
                        ])
                    filenames[newfile] = calculate_score(newfile)
    if user:
        for filename in list(filenames):
            for match in [True, False]:
                for label in CONDITION[user]['labels']:
                    newfile = filename
                    if not newfile.endswith('##'):
                        newfile += ','
                    newfile += '.'.join([
                        label,
                        local_user if match else 'baduser'
                        ])
                    filenames[newfile] = calculate_score(newfile)

    script = f"""
        YADM_TEST=1 source {yadm}
        score=0
        local_class={local_class}
        local_system={local_system}
        local_distro={local_distro}
        local_host={local_host}
        local_user={local_user}
    """
    expected = ''
    for filename in filenames:
        script += f"""
            score_file "{filename}"
            echo "{filename}"
            echo "$score"
        """
        expected += filename + '\n'
        expected += str(filenames[filename]) + '\n'
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out == expected


def test_score_values_templates(runner, yadm):
    """Test score results"""
    local_class = 'testclass'
    local_system = 'testsystem'
    local_distro = 'testdistro'
    local_host = 'testhost'
    local_user = 'testuser'
    filenames = {'filename##': 0}

    for filename in list(filenames):
        for label in TEMPLATE_LABELS:
            newfile = filename
            if not newfile.endswith('##'):
                newfile += ','
            newfile += '.'.join([label, 'testtemplate'])
            filenames[newfile] = calculate_score(newfile)

    script = f"""
        YADM_TEST=1 source {yadm}
        score=0
        local_class={local_class}
        local_system={local_system}
        local_distro={local_distro}
        local_host={local_host}
        local_user={local_user}
    """
    expected = ''
    for filename in filenames:
        script += f"""
            score_file "{filename}"
            echo "{filename}"
            echo "$score"
        """
        expected += filename + '\n'
        expected += str(filenames[filename]) + '\n'
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out == expected


@pytest.mark.parametrize(
    'cmd_generated',
    [True, False],
    ids=['supported-template', 'unsupported-template'])
def test_template_recording(runner, yadm, cmd_generated):
    """Template should be recorded if choose_template_cmd outputs a command"""

    mock = 'function choose_template_cmd() { return; }'
    expected = ''
    if cmd_generated:
        mock = 'function choose_template_cmd() { echo "test_cmd"; }'
        expected = 'template recorded'

    script = f"""
        YADM_TEST=1 source {yadm}
        function record_template() {{ echo "template recorded"; }}
        {mock}
        score_file "testfile##template.kind"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out.rstrip() == expected


def test_invalid(runner, yadm):
    """Verify invalid alternates are noted in INVALID_ALT"""

    invalid_file = "file##invalid"

    script = f"""
        YADM_TEST=1 source {yadm}
        score_file "{invalid_file}"
        echo "INVALID:${{INVALID_ALT[@]}}"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out.rstrip() == f'INVALID:{invalid_file}'



"""Test list"""

import os
import pytest


@pytest.mark.parametrize(
    'location', [
        'work',
        'outside',
        'subdir',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_list(runner, yadm_y, paths, ds1, location):
    """List tests"""
    if location == 'work':
        run_dir = paths.work
    elif location == 'outside':
        run_dir = paths.work.join('..')
    elif location == 'subdir':
        # first directory with tracked data
        run_dir = paths.work.join(ds1.tracked_dirs[0])
    with run_dir.as_cwd():
        # test with '-a'
        # should get all tracked files, relative to the work path
        run = runner(command=yadm_y('list', '-a'))
        assert run.success
        assert run.err == ''
        returned_files = set(run.out.splitlines())
        expected_files = {e.path for e in ds1 if e.tracked}
        assert returned_files == expected_files
        # test without '-a'
        # should get all tracked files, relative to the work path unless in a
        # subdir, then those should be a limited set of files, relative to the
        # subdir
        run = runner(command=yadm_y('list'))
        assert run.success
        assert run.err == ''
        returned_files = set(run.out.splitlines())
        if location == 'subdir':
            basepath = os.path.basename(os.getcwd())
            # only expect files within the subdir
            # names should be relative to subdir
            expected_files = {
                e.path[len(basepath)+1:]
                for e in ds1 if e.tracked and e.path.startswith(basepath)
            }
        assert returned_files == expected_files



"""Test perms"""

import os
import pytest


@pytest.mark.parametrize('autoperms', ['notest', 'unset', 'true', 'false'])
@pytest.mark.usefixtures('ds1_copy')
def test_perms(runner, yadm_y, paths, ds1, autoperms):
    """Test perms"""
    # set the value of auto-perms
    if autoperms != 'notest':
        if autoperms != 'unset':
            os.system(' '.join(yadm_y('config', 'yadm.auto-perms', autoperms)))

    # privatepaths will hold all paths that should become secured
    privatepaths = [paths.work.join('.ssh'), paths.work.join('.gnupg')]
    privatepaths += [paths.work.join(private.path) for private in ds1.private]

    # create an archive file
    os.system(f'touch "{str(paths.archive)}"')
    privatepaths.append(paths.archive)

    # create encrypted file test data
    efile1 = paths.work.join('efile1')
    efile1.write('efile1')
    efile2 = paths.work.join('efile2')
    efile2.write('efile2')
    paths.encrypt.write('efile1\nefile2\n!efile1\n')
    insecurepaths = [efile1]
    privatepaths.append(efile2)

    # assert these paths begin unsecured
    for private in privatepaths + insecurepaths:
        assert not oct(private.stat().mode).endswith('00'), (
            'Path started secured')

    cmd = 'perms'
    if autoperms != 'notest':
        cmd = 'status'
    run = runner(yadm_y(cmd), env={'HOME': paths.work})
    assert run.success
    assert run.err == ''
    if cmd == 'perms':
        assert run.out == ''

    # these paths should be secured if processing perms
    for private in privatepaths:
        if autoperms == 'false':
            assert not oct(private.stat().mode).endswith('00'), (
                'Path should not be secured')
        else:
            assert oct(private.stat().mode).endswith('00'), (
                'Path has not been secured')

    # these paths should never be secured
    for private in insecurepaths:
        assert not oct(private.stat().mode).endswith('00'), (
            'Path should not be secured')


@pytest.mark.parametrize('sshperms', [None, 'true', 'false'])
@pytest.mark.parametrize('gpgperms', [None, 'true', 'false'])
@pytest.mark.usefixtures('ds1_copy')
def test_perms_control(runner, yadm_y, paths, ds1, sshperms, gpgperms):
    """Test fine control of perms"""
    # set the value of ssh-perms
    if sshperms:
        os.system(' '.join(yadm_y('config', 'yadm.ssh-perms', sshperms)))

    # set the value of gpg-perms
    if gpgperms:
        os.system(' '.join(yadm_y('config', 'yadm.gpg-perms', gpgperms)))

    # privatepaths will hold all paths that should become secured
    privatepaths = [paths.work.join('.ssh'), paths.work.join('.gnupg')]
    privatepaths += [paths.work.join(private.path) for private in ds1.private]

    # assert these paths begin unsecured
    for private in privatepaths:
        assert not oct(private.stat().mode).endswith('00'), (
            'Path started secured')

    run = runner(yadm_y('perms'), env={'HOME': paths.work})
    assert run.success
    assert run.err == ''
    assert run.out == ''

    # these paths should be secured if processing perms
    for private in privatepaths:
        if (
                (sshperms == 'false' and 'ssh' in str(private))
                or
                (gpgperms == 'false' and 'gnupg' in str(private))
        ):
            assert not oct(private.stat().mode).endswith('00'), (
                'Path should not be secured')
        else:
            assert oct(private.stat().mode).endswith('00'), (
                'Path has not been secured')

    # verify permissions aren't changed for the worktree
    assert oct(paths.work.stat().mode).endswith('0755')



"""Unit tests: query_distro"""
import pytest


@pytest.mark.parametrize(
    'condition', ['lsb_release', 'os-release', 'missing'])
def test_query_distro(runner, yadm, tst_distro, tmp_path, condition):
    """Match lsb_release -si when present"""
    test_release = 'testrelease'
    lsb_release = ''
    os_release = tmp_path.joinpath('os-release')
    if condition == 'os-release':
        os_release.write_text(f"testing\nID={test_release}\nrelease")
    if condition != 'lsb_release':
        lsb_release = 'LSB_RELEASE_PROGRAM="missing_lsb_release"'
    script = f"""
        YADM_TEST=1 source {yadm}
        {lsb_release}
        OS_RELEASE="{os_release}"
        query_distro
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if condition == 'lsb_release':
        assert run.out.rstrip() == tst_distro
    elif condition == 'os-release':
        assert run.out.rstrip() == test_release
    else:
        assert run.out.rstrip() == ''



"""Unit tests: private_dirs"""
import pytest


@pytest.mark.parametrize(
    'gnupghome',
    [True, False],
    ids=['gnupghome-set', 'gnupghome-unset'],
)
@pytest.mark.parametrize('param', ['all', 'gnupg'])
def test_relative_path(runner, paths, gnupghome, param):
    """Test translate_to_relative"""

    alt_gnupghome = 'alt/gnupghome'
    env_gnupghome = paths.work.join(alt_gnupghome)

    script = f"""
        YADM_TEST=1 source {paths.pgm}
        YADM_WORK={paths.work}
        private_dirs {param}
    """

    env = {}
    if gnupghome:
        env['GNUPGHOME'] = env_gnupghome

    expected = alt_gnupghome if gnupghome else '.gnupg'
    if param == 'all':
        expected = f'.ssh {expected}'

    run = runner(command=['bash'], inp=script, env=env)
    assert run.success
    assert run.err == ''
    assert run.out.strip() == expected



"""Unit tests: is_valid_branch_name"""
import pytest

# Git branches do not allow:
#  * path component that begins with "."
#  * double dot
#  * "~", "^", ":", "\", space
#  * end with a "/"
#  * end with ".lock"


@pytest.mark.parametrize(
    'branch, expected', [
        ('master', 'valid'),
        ('path/branch', 'valid'),
        ('path/.branch', 'invalid'),
        ('path..branch', 'invalid'),
        ('path~branch', 'invalid'),
        ('path^branch', 'invalid'),
        ('path:branch', 'invalid'),
        ('path\\branch', 'invalid'),
        ('path branch', 'invalid'),
        ('path/branch/', 'invalid'),
        ('branch.lock', 'invalid'),
    ])
def test_is_valid_branch_name(runner, yadm, branch, expected):
    """Test function is_valid_branch_name()"""

    script = f"""
        YADM_TEST=1 source {yadm}
        if is_valid_branch_name "{branch}"; then
            echo valid
        else
            echo invalid
        fi
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out.strip() == expected



"""Test config"""

import os
import pytest

TEST_SECTION = 'test'
TEST_ATTRIBUTE = 'attribute'
TEST_KEY = f'{TEST_SECTION}.{TEST_ATTRIBUTE}'
TEST_VALUE = 'testvalue'
TEST_FILE = f'[{TEST_SECTION}]\n\t{TEST_ATTRIBUTE} = {TEST_VALUE}'


def test_config_no_params(runner, yadm_y, supported_configs):
    """No parameters

    Display instructions
    Display supported configs
    Exit with 0
    """

    run = runner(yadm_y('config'))

    assert run.success
    assert run.err == ''
    assert 'Please read the CONFIGURATION section' in run.out
    for config in supported_configs:
        assert config in run.out


def test_config_read_missing(runner, yadm_y):
    """Read missing attribute

    Display an empty value
    Exit with 0
    """

    run = runner(yadm_y('config', TEST_KEY))

    assert run.success
    assert run.err == ''
    assert run.out == ''


def test_config_write(runner, yadm_y, paths):
    """Write attribute

    Display no output
    Update configuration file
    Exit with 0
    """

    run = runner(yadm_y('config', TEST_KEY, TEST_VALUE))

    assert run.success
    assert run.err == ''
    assert run.out == ''
    assert paths.config.read().strip() == TEST_FILE


def test_config_read(runner, yadm_y, paths):
    """Read attribute

    Display value
    Exit with 0
    """

    paths.config.write(TEST_FILE)
    run = runner(yadm_y('config', TEST_KEY))

    assert run.success
    assert run.err == ''
    assert run.out.strip() == TEST_VALUE


def test_config_update(runner, yadm_y, paths):
    """Update attribute

    Display no output
    Update configuration file
    Exit with 0
    """

    paths.config.write(TEST_FILE)

    run = runner(yadm_y('config', TEST_KEY, TEST_VALUE + 'extra'))

    assert run.success
    assert run.err == ''
    assert run.out == ''

    assert paths.config.read().strip() == TEST_FILE + 'extra'


@pytest.mark.usefixtures('ds1_repo_copy')
def test_config_local_read(runner, yadm_y, paths, supported_local_configs):
    """Read local attribute

    Display value from the repo config
    Exit with 0
    """

    # populate test values
    for config in supported_local_configs:
        os.system(
            f'GIT_DIR="{paths.repo}" '
            f'git config --local "{config}" "value_of_{config}"')

    # run yadm config
    for config in supported_local_configs:
        run = runner(yadm_y('config', config))
        assert run.success
        assert run.err == ''
        assert run.out.strip() == f'value_of_{config}'


@pytest.mark.usefixtures('ds1_repo_copy')
def test_config_local_write(runner, yadm_y, paths, supported_local_configs):
    """Write local attribute

    Display no output
    Write value to the repo config
    Exit with 0
    """

    # run yadm config
    for config in supported_local_configs:
        run = runner(yadm_y('config', config, f'value_of_{config}'))
        assert run.success
        assert run.err == ''
        assert run.out == ''

    # verify test values
    for config in supported_local_configs:
        run = runner(
            command=('git', 'config', config),
            env={'GIT_DIR': paths.repo})
        assert run.success
        assert run.err == ''
        assert run.out.strip() == f'value_of_{config}'



"""Unit tests: parse_encrypt"""

import pytest


def test_not_called(runner, paths):
    """Test parse_encrypt (not called)"""
    run = run_parse_encrypt(runner, paths, skip_parse=True)
    assert run.success
    assert run.err == ''
    assert 'EIF:unparsed' in run.out, 'EIF should be unparsed'
    assert 'EIF_COUNT:1' in run.out, 'Only value of EIF should be unparsed'


def test_short_circuit(runner, paths):
    """Test parse_encrypt (short-circuit)"""
    run = run_parse_encrypt(runner, paths, twice=True)
    assert run.success
    assert run.err == ''
    assert 'PARSE_ENCRYPT_SHORT=parse_encrypt() not reprocessed' in run.out, (
        'parse_encrypt() should short-circuit')


@pytest.mark.parametrize(
    'encrypt', [
        ('missing'),
        ('empty'),
    ])
def test_empty(runner, paths, encrypt):
    """Test parse_encrypt (file missing/empty)"""

    # write encrypt file
    if encrypt == 'missing':
        assert not paths.encrypt.exists(), 'Encrypt should be missing'
    else:
        paths.encrypt.write('')
        assert paths.encrypt.exists(), 'Encrypt should exist'
        assert paths.encrypt.size() == 0, 'Encrypt should be empty'

    # run parse_encrypt
    run = run_parse_encrypt(runner, paths)
    assert run.success
    assert run.err == ''

    # validate parsing result
    assert 'EIF_COUNT:0' in run.out, 'EIF should be empty'


def create_test_encrypt_data(paths):
    """Generate test data for testing encrypt"""

    edata = ''
    expected = set()

    # empty line
    edata += '\n'

    # simple comments
    edata += '# a simple comment\n'
    edata += '    # a comment with leading space\n'

    # unreferenced directory
    paths.work.join('unreferenced').mkdir()

    # simple files
    edata += 'simple_file\n'
    edata += 'simple.file\n'
    paths.work.join('simple_file').write('')
    paths.work.join('simple.file').write('')
    paths.work.join('simple_file2').write('')
    paths.work.join('simple.file2').write('')
    expected.add('simple_file')
    expected.add('simple.file')

    # simple files in directories
    edata += 'simple_dir/simple_file\n'
    paths.work.join('simple_dir/simple_file').write('', ensure=True)
    paths.work.join('simple_dir/simple_file2').write('', ensure=True)
    expected.add('simple_dir/simple_file')

    # paths with spaces
    edata += 'with space/with space\n'
    paths.work.join('with space/with space').write('', ensure=True)
    paths.work.join('with space/with space2').write('', ensure=True)
    expected.add('with space/with space')

    # hidden files
    edata += '.hidden\n'
    paths.work.join('.hidden').write('')
    expected.add('.hidden')

    # hidden files in directories
    edata += '.hidden_dir/.hidden_file\n'
    paths.work.join('.hidden_dir/.hidden_file').write('', ensure=True)
    expected.add('.hidden_dir/.hidden_file')

    # wildcards
    edata += 'wild*\n'
    paths.work.join('wildcard1').write('', ensure=True)
    paths.work.join('wildcard2').write('', ensure=True)
    expected.add('wildcard1')
    expected.add('wildcard2')

    edata += 'dirwild*\n'
    paths.work.join('dirwildcard/file1').write('', ensure=True)
    paths.work.join('dirwildcard/file2').write('', ensure=True)
    expected.add('dirwildcard')

    # excludes
    edata += 'exclude*\n'
    edata += 'ex ex/*\n'
    paths.work.join('exclude_file1').write('')
    paths.work.join('exclude_file2.ex').write('')
    paths.work.join('exclude_file3.ex3').write('')
    expected.add('exclude_file1')
    expected.add('exclude_file3.ex3')
    edata += '!*.ex\n'
    edata += '!ex ex/*.txt\n'
    paths.work.join('ex ex/file4').write('', ensure=True)
    paths.work.join('ex ex/file5.txt').write('', ensure=True)
    paths.work.join('ex ex/file6.text').write('', ensure=True)
    expected.add('ex ex/file4')
    expected.add('ex ex/file6.text')

    # double star
    edata += 'doublestar/**/file*\n'
    edata += '!**/file3\n'
    paths.work.join('doublestar/a/b/file1').write('', ensure=True)
    paths.work.join('doublestar/c/d/file2').write('', ensure=True)
    paths.work.join('doublestar/e/f/file3').write('', ensure=True)
    paths.work.join('doublestar/g/h/nomatch').write('', ensure=True)
    expected.add('doublestar/a/b/file1')
    expected.add('doublestar/c/d/file2')
    # doublestar/e/f/file3 is excluded

    return edata, expected


@pytest.mark.usefixtures('ds1_repo_copy')
def test_file_parse_encrypt(runner, paths):
    """Test parse_encrypt

    Test an array of supported features of the encrypt configuration.
    """

    # generate test data & expectations
    edata, expected = create_test_encrypt_data(paths)

    # write encrypt file
    print(f'ENCRYPT:\n---\n{edata}---\n')
    paths.encrypt.write(edata)
    assert paths.encrypt.isfile()

    # run parse_encrypt
    run = run_parse_encrypt(runner, paths)
    assert run.success
    assert run.err == ''

    assert f'EIF_COUNT:{len(expected)}' in run.out, 'EIF count wrong'
    for expected_file in expected:
        assert f'EIF:{expected_file}\n' in run.out

    sorted_expectations = '\n'.join(
        [f'EIF:{exp}' for exp in sorted(expected)])
    assert sorted_expectations in run.out


def run_parse_encrypt(
        runner, paths,
        skip_parse=False,
        twice=False):
    """Run parse_encrypt

    A count of ENCRYPT_INCLUDE_FILES will be reported as EIF_COUNT:X. All
    values of ENCRYPT_INCLUDE_FILES will be reported as individual EIF:value
    lines.
    """
    parse_cmd = 'parse_encrypt'
    if skip_parse:
        parse_cmd = ''
    if twice:
        parse_cmd = 'parse_encrypt; parse_encrypt'
    script = f"""
        YADM_TEST=1 source {paths.pgm}
        YADM_ENCRYPT={paths.encrypt}
        export YADM_ENCRYPT
        GIT_DIR={paths.repo}
        export GIT_DIR
        YADM_WORK={paths.work}
        export YADM_WORK
        {parse_cmd}
        export ENCRYPT_INCLUDE_FILES
        export PARSE_ENCRYPT_SHORT
        env
        echo EIF_COUNT:${{#ENCRYPT_INCLUDE_FILES[@]}}
        for value in "${{ENCRYPT_INCLUDE_FILES[@]}}"; do
            echo "EIF:$value"
        done
    """
    run = runner(command=['bash'], inp=script)
    return run



"""Unit tests: bootstrap_available"""


def test_bootstrap_missing(runner, paths):
    """Test result of bootstrap_available, when bootstrap missing"""
    run_test(runner, paths, False)


def test_bootstrap_no_exec(runner, paths):
    """Test result of bootstrap_available, when bootstrap not executable"""
    paths.bootstrap.write('')
    paths.bootstrap.chmod(0o644)
    run_test(runner, paths, False)


def test_bootstrap_exec(runner, paths):
    """Test result of bootstrap_available, when bootstrap executable"""
    paths.bootstrap.write('')
    paths.bootstrap.chmod(0o775)
    run_test(runner, paths, True)


def run_test(runner, paths, success):
    """Run bootstrap_available, and test result"""
    script = f"""
        YADM_TEST=1 source {paths.pgm}
        YADM_BOOTSTRAP='{paths.bootstrap}'
        bootstrap_available
    """
    run = runner(command=['bash'], inp=script)
    assert run.success == success
    assert run.err == ''
    assert run.out == ''



"""Unit tests: template_j2cli & template_envtpl"""
import pytest

LOCAL_CLASS = "j2_Test+@-!^Class"
LOCAL_SYSTEM = "j2_Test+@-!^System"
LOCAL_HOST = "j2_Test+@-!^Host"
LOCAL_USER = "j2_Test+@-!^User"
LOCAL_DISTRO = "j2_Test+@-!^Distro"
TEMPLATE = f'''
start of template
j2 class  = >{{{{YADM_CLASS}}}}<
j2 os     = >{{{{YADM_OS}}}}<
j2 host   = >{{{{YADM_HOSTNAME}}}}<
j2 user   = >{{{{YADM_USER}}}}<
j2 distro = >{{{{YADM_DISTRO}}}}<
{{%- if YADM_CLASS == "wrongclass1" %}}
wrong class 1
{{%- endif %}}
{{%- if YADM_CLASS == "{LOCAL_CLASS}" %}}
Included section for class = {{{{YADM_CLASS}}}} ({{{{YADM_CLASS}}}} repeated)
{{%- endif %}}
{{%- if YADM_CLASS == "wrongclass2" %}}
wrong class 2
{{%- endif %}}
{{%- if YADM_OS == "wrongos1" %}}
wrong os 1
{{%- endif %}}
{{%- if YADM_OS == "{LOCAL_SYSTEM}" %}}
Included section for os = {{{{YADM_OS}}}} ({{{{YADM_OS}}}} repeated)
{{%- endif %}}
{{%- if YADM_OS == "wrongos2" %}}
wrong os 2
{{%- endif %}}
{{%- if YADM_HOSTNAME == "wronghost1" %}}
wrong host 1
{{%- endif %}}
{{%- if YADM_HOSTNAME == "{LOCAL_HOST}" %}}
Included section for host = {{{{YADM_HOSTNAME}}}} ({{{{YADM_HOSTNAME}}}} again)
{{%- endif %}}
{{%- if YADM_HOSTNAME == "wronghost2" %}}
wrong host 2
{{%- endif %}}
{{%- if YADM_USER == "wronguser1" %}}
wrong user 1
{{%- endif %}}
{{%- if YADM_USER == "{LOCAL_USER}" %}}
Included section for user = {{{{YADM_USER}}}} ({{{{YADM_USER}}}} repeated)
{{%- endif %}}
{{%- if YADM_USER == "wronguser2" %}}
wrong user 2
{{%- endif %}}
{{%- if YADM_DISTRO == "wrongdistro1" %}}
wrong distro 1
{{%- endif %}}
{{%- if YADM_DISTRO == "{LOCAL_DISTRO}" %}}
Included section for distro = {{{{YADM_DISTRO}}}} ({{{{YADM_DISTRO}}}} again)
{{%- endif %}}
{{%- if YADM_DISTRO == "wrongdistro2" %}}
wrong distro 2
{{%- endif %}}
end of template
'''
EXPECTED = f'''
start of template
j2 class  = >{LOCAL_CLASS}<
j2 os     = >{LOCAL_SYSTEM}<
j2 host   = >{LOCAL_HOST}<
j2 user   = >{LOCAL_USER}<
j2 distro = >{LOCAL_DISTRO}<
Included section for class = {LOCAL_CLASS} ({LOCAL_CLASS} repeated)
Included section for os = {LOCAL_SYSTEM} ({LOCAL_SYSTEM} repeated)
Included section for host = {LOCAL_HOST} ({LOCAL_HOST} again)
Included section for user = {LOCAL_USER} ({LOCAL_USER} repeated)
Included section for distro = {LOCAL_DISTRO} ({LOCAL_DISTRO} again)
end of template
'''


@pytest.mark.parametrize('processor', ('j2cli', 'envtpl'))
def test_template_j2(runner, yadm, tmpdir, processor):
    """Test processing by j2cli & envtpl"""

    input_file = tmpdir.join('input')
    input_file.write(TEMPLATE, ensure=True)
    output_file = tmpdir.join('output')

    script = f"""
        YADM_TEST=1 source {yadm}
        local_class="{LOCAL_CLASS}"
        local_system="{LOCAL_SYSTEM}"
        local_host="{LOCAL_HOST}"
        local_user="{LOCAL_USER}"
        local_distro="{LOCAL_DISTRO}"
        template_{processor} "{input_file}" "{output_file}"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert output_file.read() == EXPECTED


@pytest.mark.parametrize('processor', ('j2cli', 'envtpl'))
def test_source(runner, yadm, tmpdir, processor):
    """Test YADM_SOURCE"""

    input_file = tmpdir.join('input')
    input_file.write('{{YADM_SOURCE}}', ensure=True)
    output_file = tmpdir.join('output')

    script = f"""
        YADM_TEST=1 source {yadm}
        template_{processor} "{input_file}" "{output_file}"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert output_file.read().strip() == str(input_file)



"""Unit tests: relative_path"""
import pytest


@pytest.mark.parametrize(
    'base,full_path,expected',
    [
        ("/A/B/C", "/A", "../.."),
        ("/A/B/C", "/A/B", ".."),
        ("/A/B/C", "/A/B/C", ""),
        ("/A/B/C", "/A/B/C/D", "D"),
        ("/A/B/C", "/A/B/C/D/E", "D/E"),
        ("/A/B/C", "/A/B/D", "../D"),
        ("/A/B/C", "/A/B/D/E", "../D/E"),
        ("/A/B/C", "/A/D", "../../D"),
        ("/A/B/C", "/A/D/E", "../../D/E"),
        ("/A/B/C", "/D/E/F", "../../../D/E/F"),
    ],
)
def test_relative_path(runner, paths, base, full_path, expected):
    """Test translate_to_relative"""

    script = f"""
        YADM_TEST=1 source {paths.pgm}
        relative_path "{base}" "{full_path}"
    """

    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert run.out.strip() == expected



"""Test yadm.alt-copy"""

import os
import pytest


@pytest.mark.parametrize(
    'cygwin',
    [pytest.param(True, marks=pytest.mark.deprecated), False],
    ids=['cygwin', 'no-cygwin'])
@pytest.mark.parametrize(
    'setting, expect_link, pre_existing', [
        (None, True, None),
        (True, False, None),
        (False, True, None),
        (True, False, 'link'),
        (True, False, 'file'),
    ],
    ids=[
        'unset',
        'true',
        'false',
        'pre-existing symlink',
        'pre-existing file',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_alt_copy(
        runner, yadm_y, paths, tst_sys,
        setting, expect_link, pre_existing,
        cygwin):
    """Test yadm.alt-copy"""

    option = 'yadm.cygwin-copy' if cygwin else 'yadm.alt-copy'

    if setting is not None:
        os.system(' '.join(yadm_y('config', option, str(setting))))

    expected_content = f'test_alt_copy##os.{tst_sys}'

    alt_path = paths.work.join('test_alt_copy')
    if pre_existing == 'symlink':
        alt_path.mklinkto(expected_content)
    elif pre_existing == 'file':
        alt_path.write('wrong content')

    run = runner(yadm_y('alt'))
    assert run.success
    assert run.err == ''
    assert 'Linking' in run.out

    assert alt_path.read() == expected_content
    assert alt_path.islink() == expect_link



"""Unit tests: template_default"""

# these values are also testing the handling of bizarre characters
LOCAL_CLASS = "default_Test+@-!^Class"
LOCAL_SYSTEM = "default_Test+@-!^System"
LOCAL_HOST = "default_Test+@-!^Host"
LOCAL_USER = "default_Test+@-!^User"
LOCAL_DISTRO = "default_Test+@-!^Distro"
TEMPLATE = f'''
start of template
default class  = >{{{{yadm.class}}}}<
default os     = >{{{{yadm.os}}}}<
default host   = >{{{{yadm.hostname}}}}<
default user   = >{{{{yadm.user}}}}<
default distro = >{{{{yadm.distro}}}}<
{{% if yadm.class == "else1" %}}
wrong else 1
{{% else %}}
Included section from else
{{% endif %}}
{{% if yadm.class == "wrongclass1" %}}
wrong class 1
{{% endif %}}
{{% if yadm.class == "{LOCAL_CLASS}" %}}
Included section for class = {{{{yadm.class}}}} ({{{{yadm.class}}}} repeated)
Multiple lines
{{% else %}}
Should not be included...
{{% endif %}}
{{% if yadm.class == "wrongclass2" %}}
wrong class 2
{{% endif %}}
{{% if yadm.os == "wrongos1" %}}
wrong os 1
{{% endif %}}
{{% if yadm.os == "{LOCAL_SYSTEM}" %}}
Included section for os = {{{{yadm.os}}}} ({{{{yadm.os}}}} repeated)
{{% endif %}}
{{% if yadm.os == "wrongos2" %}}
wrong os 2
{{% endif %}}
{{% if yadm.hostname == "wronghost1" %}}
wrong host 1
{{% endif %}}
{{% if yadm.hostname == "{LOCAL_HOST}" %}}
Included section for host = {{{{yadm.hostname}}}} ({{{{yadm.hostname}}}} again)
{{% endif %}}
{{% if yadm.hostname == "wronghost2" %}}
wrong host 2
{{% endif %}}
{{% if yadm.user == "wronguser1" %}}
wrong user 1
{{% endif %}}
{{% if yadm.user == "{LOCAL_USER}" %}}
Included section for user = {{{{yadm.user}}}} ({{{{yadm.user}}}} repeated)
{{% endif %}}
{{% if yadm.user == "wronguser2" %}}
wrong user 2
{{% endif %}}
{{% if yadm.distro == "wrongdistro1" %}}
wrong distro 1
{{% endif %}}
{{% if yadm.distro == "{LOCAL_DISTRO}" %}}
Included section for distro = {{{{yadm.distro}}}} ({{{{yadm.distro}}}} again)
{{% endif %}}
{{% if yadm.distro == "wrongdistro2" %}}
wrong distro 2
{{% endif %}}
end of template
'''
EXPECTED = f'''
start of template
default class  = >{LOCAL_CLASS}<
default os     = >{LOCAL_SYSTEM}<
default host   = >{LOCAL_HOST}<
default user   = >{LOCAL_USER}<
default distro = >{LOCAL_DISTRO}<
Included section from else
Included section for class = {LOCAL_CLASS} ({LOCAL_CLASS} repeated)
Multiple lines
Included section for os = {LOCAL_SYSTEM} ({LOCAL_SYSTEM} repeated)
Included section for host = {LOCAL_HOST} ({LOCAL_HOST} again)
Included section for user = {LOCAL_USER} ({LOCAL_USER} repeated)
Included section for distro = {LOCAL_DISTRO} ({LOCAL_DISTRO} again)
end of template
'''


def test_template_default(runner, yadm, tmpdir):
    """Test template_default"""

    input_file = tmpdir.join('input')
    input_file.write(TEMPLATE, ensure=True)
    output_file = tmpdir.join('output')

    script = f"""
        YADM_TEST=1 source {yadm}
        set_awk
        local_class="{LOCAL_CLASS}"
        local_system="{LOCAL_SYSTEM}"
        local_host="{LOCAL_HOST}"
        local_user="{LOCAL_USER}"
        local_distro="{LOCAL_DISTRO}"
        template_default "{input_file}" "{output_file}"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert output_file.read() == EXPECTED


def test_source(runner, yadm, tmpdir):
    """Test yadm.source"""

    input_file = tmpdir.join('input')
    input_file.write('{{yadm.source}}', ensure=True)
    output_file = tmpdir.join('output')

    script = f"""
        YADM_TEST=1 source {yadm}
        set_awk
        template_default "{input_file}" "{output_file}"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert output_file.read().strip() == str(input_file)



"""Syntax checks"""

import os
import pytest


def test_yadm_syntax(runner, yadm):
    """Is syntactically valid"""
    run = runner(command=['bash', '-n', yadm])
    assert run.success


def test_shellcheck(pytestconfig, runner, yadm, shellcheck_version):
    """Passes shellcheck"""
    if not pytestconfig.getoption("--force-linters"):
        run = runner(command=['shellcheck', '-V'], report=False)
        if f'version: {shellcheck_version}' not in run.out:
            pytest.skip('Unsupported shellcheck version')
    run = runner(command=['shellcheck', '-s', 'bash', yadm])
    assert run.success


def test_pylint(pytestconfig, runner, pylint_version):
    """Passes pylint"""
    if not pytestconfig.getoption("--force-linters"):
        run = runner(command=['pylint', '--version'], report=False)
        if f'pylint {pylint_version}' not in run.out:
            pytest.skip('Unsupported pylint version')
    pyfiles = list()
    for tfile in os.listdir('test'):
        if tfile.endswith('.py'):
            pyfiles.append(f'test/{tfile}')
    run = runner(command=['pylint'] + pyfiles)
    assert run.success


def test_flake8(pytestconfig, runner, flake8_version):
    """Passes flake8"""
    if not pytestconfig.getoption("--force-linters"):
        run = runner(command=['flake8', '--version'], report=False)
        if not run.out.startswith(flake8_version):
            pytest.skip('Unsupported flake8 version')
    run = runner(command=['flake8', 'test'])
    assert run.success


def test_yamllint(pytestconfig, runner, yamllint_version):
    """Passes yamllint"""
    if not pytestconfig.getoption("--force-linters"):
        run = runner(command=['yamllint', '--version'], report=False)
        if not run.out.strip().endswith(yamllint_version):
            pytest.skip('Unsupported yamllint version')
    run = runner(
        command=['yamllint', '-s', '$(find . -name \\*.yml)'],
        shell=True)
    assert run.success



"""Test alt"""

import os
import string
import py
import pytest
import utils

# These tests are for the alternate processing in YADM_COMPATIBILITY=1 mode
pytestmark = pytest.mark.deprecated

# These test IDs are broken. During the writing of these tests, problems have
# been discovered in the way yadm orders matching files.
BROKEN_TEST_IDS = [
    'test_wild[tracked-##C.S.H.U-C-S%-H%-U]',
    'test_wild[tracked-##C.S.H.U-C-S-H%-U]',
    'test_wild[encrypted-##C.S.H.U-C-S%-H%-U]',
    'test_wild[encrypted-##C.S.H.U-C-S-H%-U]',
    ]

PRECEDENCE = [
    '##',
    '##$tst_sys',
    '##$tst_sys.$tst_host',
    '##$tst_sys.$tst_host.$tst_user',
    '##$tst_class',
    '##$tst_class.$tst_sys',
    '##$tst_class.$tst_sys.$tst_host',
    '##$tst_class.$tst_sys.$tst_host.$tst_user',
    ]

WILD_TEMPLATES = [
    '##$tst_class',
    '##$tst_class.$tst_sys',
    '##$tst_class.$tst_sys.$tst_host',
    '##$tst_class.$tst_sys.$tst_host.$tst_user',
    ]

TEST_PATHS = [utils.ALT_FILE1, utils.ALT_FILE2, utils.ALT_DIR]

WILD_TESTED = set()


@pytest.mark.parametrize('precedence_index', range(len(PRECEDENCE)))
@pytest.mark.parametrize(
    'tracked, encrypt, exclude', [
        (False, False, False),
        (True, False, False),
        (False, True, False),
        (False, True, True),
    ], ids=[
        'untracked',
        'tracked',
        'encrypted',
        'excluded',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_alt(runner, yadm_y, paths,
             tst_sys, tst_host, tst_user,
             tracked, encrypt, exclude,
             precedence_index):
    """Test alternate linking

    This test is done by iterating for the number of templates in PRECEDENCE.
    With each iteration, another file is left off the list. So with each
    iteration, the template with the "highest precedence" is left out. The file
    using the highest precedence should be the one linked.
    """

    # set the class
    tst_class = 'testclass'
    utils.set_local(paths, 'class', tst_class)

    # process the templates in PRECEDENCE
    precedence = list()
    for template in PRECEDENCE:
        precedence.append(
            string.Template(template).substitute(
                tst_class=tst_class,
                tst_host=tst_host,
                tst_sys=tst_sys,
                tst_user=tst_user,
            )
        )

    # create files using a subset of files
    for suffix in precedence[0:precedence_index+1]:
        utils.create_alt_files(paths, suffix, tracked=tracked,
                               encrypt=encrypt, exclude=exclude)

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + precedence[precedence_index]
        if tracked or (encrypt and not exclude):
            assert paths.work.join(file_path).islink()
            target = py.path.local(
                os.path.realpath(paths.work.join(file_path)))
            if target.isfile():
                assert paths.work.join(file_path).read() == source_file
                assert str(paths.work.join(source_file)) in linked
            else:
                assert paths.work.join(file_path).join(
                    utils.CONTAINED).read() == source_file
                assert str(paths.work.join(source_file)) in linked
        else:
            assert not paths.work.join(file_path).exists()
            assert str(paths.work.join(source_file)) not in linked


def short_template(template):
    """Translate template into something short for test IDs"""
    return string.Template(template).substitute(
        tst_class='C',
        tst_host='H',
        tst_sys='S',
        tst_user='U',
    )


@pytest.mark.parametrize('wild_user', [True, False], ids=['U%', 'U'])
@pytest.mark.parametrize('wild_host', [True, False], ids=['H%', 'H'])
@pytest.mark.parametrize('wild_sys', [True, False], ids=['S%', 'S'])
@pytest.mark.parametrize('wild_class', [True, False], ids=['C%', 'C'])
@pytest.mark.parametrize('template', WILD_TEMPLATES, ids=short_template)
@pytest.mark.parametrize(
    'tracked, encrypt', [
        (True, False),
        (False, True),
    ], ids=[
        'tracked',
        'encrypted',
    ])
@pytest.mark.usefixtures('ds1_copy')
def test_wild(request, runner, yadm_y, paths,
              tst_sys, tst_host, tst_user,
              tracked, encrypt,
              wild_class, wild_host, wild_sys, wild_user,
              template):
    """Test wild linking

    These tests are done by creating permutations of the possible files using
    WILD_TEMPLATES. Each case is then tested (while skipping the already tested
    permutations for efficiency).
    """

    if request.node.name in BROKEN_TEST_IDS:
        pytest.xfail(
            'This test is known to be broken. '
            'This bug only affects deprecated features.')

    tst_class = 'testclass'

    # determine the "wild" version of the suffix
    str_class = '%' if wild_class else tst_class
    str_host = '%' if wild_host else tst_host
    str_sys = '%' if wild_sys else tst_sys
    str_user = '%' if wild_user else tst_user
    wild_suffix = string.Template(template).substitute(
        tst_class=str_class,
        tst_host=str_host,
        tst_sys=str_sys,
        tst_user=str_user,
    )

    # determine the "standard" version of the suffix
    std_suffix = string.Template(template).substitute(
        tst_class=tst_class,
        tst_host=tst_host,
        tst_sys=tst_sys,
        tst_user=tst_user,
    )

    # skip over duplicate tests (this seems to be the simplest way to cover the
    # permutations of tests, while skipping duplicates.)
    test_key = f'{tracked}{encrypt}{wild_suffix}{std_suffix}'
    if test_key in WILD_TESTED:
        return
    WILD_TESTED.add(test_key)

    # set the class
    utils.set_local(paths, 'class', tst_class)

    # create files using the wild suffix
    utils.create_alt_files(paths, wild_suffix, tracked=tracked,
                           encrypt=encrypt, exclude=False)

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + wild_suffix
        assert paths.work.join(file_path).islink()
        target = py.path.local(os.path.realpath(paths.work.join(file_path)))
        if target.isfile():
            assert paths.work.join(file_path).read() == source_file
            assert str(paths.work.join(source_file)) in linked
        else:
            assert paths.work.join(file_path).join(
                utils.CONTAINED).read() == source_file
            assert str(paths.work.join(source_file)) in linked

    # create files using the standard suffix
    utils.create_alt_files(paths, std_suffix, tracked=tracked,
                           encrypt=encrypt, exclude=False)

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + std_suffix
        assert paths.work.join(file_path).islink()
        target = py.path.local(os.path.realpath(paths.work.join(file_path)))
        if target.isfile():
            assert paths.work.join(file_path).read() == source_file
            assert str(paths.work.join(source_file)) in linked
        else:
            assert paths.work.join(file_path).join(
                utils.CONTAINED).read() == source_file
            assert str(paths.work.join(source_file)) in linked


@pytest.mark.usefixtures('ds1_copy')
def test_local_override(runner, yadm_y, paths,
                        tst_sys, tst_host, tst_user):
    """Test local overrides"""

    # define local overrides
    utils.set_local(paths, 'class', 'or-class')
    utils.set_local(paths, 'hostname', 'or-hostname')
    utils.set_local(paths, 'os', 'or-os')
    utils.set_local(paths, 'user', 'or-user')

    # create files, the first would normally be the most specific version
    # however, the second is the overridden version which should be preferred.
    utils.create_alt_files(
        paths, f'##or-class.{tst_sys}.{tst_host}.{tst_user}')
    utils.create_alt_files(
        paths, '##or-class.or-os.or-hostname.or-user')

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + '##or-class.or-os.or-hostname.or-user'
        assert paths.work.join(file_path).islink()
        target = py.path.local(os.path.realpath(paths.work.join(file_path)))
        if target.isfile():
            assert paths.work.join(file_path).read() == source_file
            assert str(paths.work.join(source_file)) in linked
        else:
            assert paths.work.join(file_path).join(
                utils.CONTAINED).read() == source_file
            assert str(paths.work.join(source_file)) in linked


@pytest.mark.parametrize('suffix', ['AAA', 'ZZZ', 'aaa', 'zzz'])
@pytest.mark.usefixtures('ds1_copy')
def test_class_case(runner, yadm_y, paths, tst_sys, suffix):
    """Test range of class cases"""

    # set the class
    utils.set_local(paths, 'class', suffix)

    # create files
    endings = [suffix]
    if tst_sys == 'Linux':
        # Only create all of these side-by-side on Linux, which is
        # unquestionably case-sensitive. This would break tests on
        # case-insensitive systems.
        endings = ['AAA', 'ZZZ', 'aaa', 'zzz']
    for ending in endings:
        utils.create_alt_files(paths, f'##{ending}')

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + f'##{suffix}'
        assert paths.work.join(file_path).islink()
        target = py.path.local(os.path.realpath(paths.work.join(file_path)))
        if target.isfile():
            assert paths.work.join(file_path).read() == source_file
            assert str(paths.work.join(source_file)) in linked
        else:
            assert paths.work.join(file_path).join(
                utils.CONTAINED).read() == source_file
            assert str(paths.work.join(source_file)) in linked


@pytest.mark.parametrize('autoalt', [None, 'true', 'false'])
@pytest.mark.usefixtures('ds1_copy')
def test_auto_alt(runner, yadm_y, paths, autoalt):
    """Test setting auto-alt"""

    # set the value of auto-alt
    if autoalt:
        os.system(' '.join(yadm_y('config', 'yadm.auto-alt', autoalt)))

    # create file
    suffix = '##'
    utils.create_alt_files(paths, suffix)

    # run status to possibly trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('status'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + suffix
        if autoalt == 'false':
            assert not paths.work.join(file_path).exists()
        else:
            assert paths.work.join(file_path).islink()
            target = py.path.local(
                os.path.realpath(paths.work.join(file_path)))
            if target.isfile():
                assert paths.work.join(file_path).read() == source_file
                # no linking output when run via auto-alt
                assert str(paths.work.join(source_file)) not in linked
            else:
                assert paths.work.join(file_path).join(
                    utils.CONTAINED).read() == source_file
                # no linking output when run via auto-alt
                assert str(paths.work.join(source_file)) not in linked


@pytest.mark.parametrize('delimiter', ['.', '_'])
@pytest.mark.usefixtures('ds1_copy')
def test_delimiter(runner, yadm_y, paths,
                   tst_sys, tst_host, tst_user, delimiter):
    """Test delimiters used"""

    suffix = '##' + delimiter.join([tst_sys, tst_host, tst_user])

    # create file
    utils.create_alt_files(paths, suffix)

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    # only a delimiter of '.' is valid
    for file_path in TEST_PATHS:
        source_file = file_path + suffix
        if delimiter == '.':
            assert paths.work.join(file_path).islink()
            target = py.path.local(
                os.path.realpath(paths.work.join(file_path)))
            if target.isfile():
                assert paths.work.join(file_path).read() == source_file
                assert str(paths.work.join(source_file)) in linked
            else:
                assert paths.work.join(file_path).join(
                    utils.CONTAINED).read() == source_file
                assert str(paths.work.join(source_file)) in linked
        else:
            assert not paths.work.join(file_path).exists()
            assert str(paths.work.join(source_file)) not in linked


@pytest.mark.usefixtures('ds1_copy')
def test_invalid_links_removed(runner, yadm_y, paths):
    """Links to invalid alternative files are removed

    This test ensures that when an already linked alternative becomes invalid
    due to a change in class, the alternate link is removed.
    """

    # set the class
    tst_class = 'testclass'
    utils.set_local(paths, 'class', tst_class)

    # create files which match the test class
    utils.create_alt_files(paths, f'##{tst_class}')

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the proper linking has occurred
    for file_path in TEST_PATHS:
        source_file = file_path + '##' + tst_class
        assert paths.work.join(file_path).islink()
        target = py.path.local(os.path.realpath(paths.work.join(file_path)))
        if target.isfile():
            assert paths.work.join(file_path).read() == source_file
            assert str(paths.work.join(source_file)) in linked
        else:
            assert paths.work.join(file_path).join(
                utils.CONTAINED).read() == source_file
            assert str(paths.work.join(source_file)) in linked

    # change the class so there are no valid alternates
    utils.set_local(paths, 'class', 'changedclass')

    # run alt to trigger linking
    env = os.environ.copy()
    env['YADM_COMPATIBILITY'] = '1'
    run = runner(yadm_y('alt'), env=env)
    assert run.success
    assert run.err == ''
    linked = utils.parse_alt_output(run.out)

    # assert the linking is removed
    for file_path in TEST_PATHS:
        source_file = file_path + '##' + tst_class
        assert not paths.work.join(file_path).exists()
        assert str(paths.work.join(source_file)) not in linked



"""Unit tests: configure_paths"""

import pytest

ARCHIVE = 'files.gpg'
BOOTSTRAP = 'bootstrap'
CONFIG = 'config'
ENCRYPT = 'encrypt'
HOME = '/testhome'
REPO = 'repo.git'
YDIR = '.config/yadm'


@pytest.mark.parametrize(
    'override, expect', [
        (None, {}),
        ('-Y', {}),
        ('--yadm-repo', {'repo': 'YADM_REPO', 'git': 'GIT_DIR'}),
        ('--yadm-config', {'config': 'YADM_CONFIG'}),
        ('--yadm-encrypt', {'encrypt': 'YADM_ENCRYPT'}),
        ('--yadm-archive', {'archive': 'YADM_ARCHIVE'}),
        ('--yadm-bootstrap', {'bootstrap': 'YADM_BOOTSTRAP'}),
    ], ids=[
        'default',
        'override yadm dir',
        'override repo',
        'override config',
        'override encrypt',
        'override archive',
        'override bootstrap',
    ])
def test_config(runner, paths, override, expect):
    """Test configure_paths"""
    opath = 'override'
    matches = match_map()
    args = []
    if override == '-Y':
        matches = match_map('/' + opath)

    if override:
        args = [override, '/' + opath]
        for ekey in expect.keys():
            matches[ekey] = f'{expect[ekey]}="/{opath}"'
        run_test(
            runner, paths,
            [override, opath],
            ['must specify a fully qualified'], 1)

    run_test(runner, paths, args, matches.values(), 0)


def match_map(yadm_dir=None):
    """Create a dictionary of matches, relative to yadm_dir"""
    if not yadm_dir:
        yadm_dir = '/'.join([HOME, YDIR])
    return {
        'yadm': f'YADM_DIR="{yadm_dir}"',
        'repo': f'YADM_REPO="{yadm_dir}/{REPO}"',
        'config': f'YADM_CONFIG="{yadm_dir}/{CONFIG}"',
        'encrypt': f'YADM_ENCRYPT="{yadm_dir}/{ENCRYPT}"',
        'archive': f'YADM_ARCHIVE="{yadm_dir}/{ARCHIVE}"',
        'bootstrap': f'YADM_BOOTSTRAP="{yadm_dir}/{BOOTSTRAP}"',
        'git': f'GIT_DIR="{yadm_dir}/{REPO}"',
        }


def run_test(runner, paths, args, expected_matches, expected_code=0):
    """Run proces global args, and run configure_paths"""
    argstring = ' '.join(['"'+a+'"' for a in args])
    script = f"""
        YADM_TEST=1 HOME="{HOME}" source {paths.pgm}
        process_global_args {argstring}
        HOME="{HOME}" set_yadm_dir
        configure_paths
        declare -p | grep -E '(YADM|GIT)_'
    """
    run = runner(command=['bash'], inp=script)
    assert run.code == expected_code
    assert run.err == ''
    for match in expected_matches:
        assert match in run.out



"""Test hooks"""

import pytest


@pytest.mark.parametrize(
    'pre, pre_code, post, post_code', [
        (False, 0, False, 0),
        (True, 0, False, 0),
        (True, 5, False, 0),
        (False, 0, True, 0),
        (False, 0, True, 5),
        (True, 0, True, 0),
        (True, 5, True, 5),
    ], ids=[
        'no-hooks',
        'pre-success',
        'pre-fail',
        'post-success',
        'post-fail',
        'pre-post-success',
        'pre-post-fail',
    ])
def test_hooks(
        runner, yadm_y, paths,
        pre, pre_code, post, post_code):
    """Test pre/post hook"""

    # generate hooks
    if pre:
        create_hook(paths, 'pre_version', pre_code)
    if post:
        create_hook(paths, 'post_version', post_code)

    # run yadm
    run = runner(yadm_y('version'))
    # when a pre hook fails, yadm should exit with the hook's code
    assert run.code == pre_code
    assert run.err == ''

    if pre:
        assert 'HOOK:pre_version' in run.out
    # if pre hook is missing or successful, yadm itself should exit 0
    if run.success:
        if post:
            assert 'HOOK:post_version' in run.out
    else:
        # when a pre hook fails, yadm should not run the command
        assert 'version will not be run' in run.out
        # when a pre hook fails, yadm should not run the post hook
        assert 'HOOK:post_version' not in run.out


# repo fixture is needed to test the population of YADM_HOOK_WORK
@pytest.mark.usefixtures('ds1_repo_copy')
def test_hook_env(runner, yadm_y, paths):
    """Test hook environment"""

    # test will be done with a non existent "git" passthru command
    # which should exit with a failing code
    cmd = 'passthrucmd'

    # write the hook
    hook = paths.hooks.join(f'post_{cmd}')
    hook.write('#!/bin/sh\nenv\n')
    hook.chmod(0o755)

    run = runner(yadm_y(cmd, 'extra_args'))

    # expect passthru to fail
    assert run.failure
    assert f"'{cmd}' is not a git command" in run.err

    # verify hook environment
    assert 'YADM_HOOK_EXIT=1\n' in run.out
    assert f'YADM_HOOK_COMMAND={cmd}\n' in run.out
    assert f'YADM_HOOK_FULL_COMMAND={cmd} extra_args\n' in run.out
    assert f'YADM_HOOK_REPO={paths.repo}\n' in run.out
    assert f'YADM_HOOK_WORK={paths.work}\n' in run.out


def create_hook(paths, name, code):
    """Create hook"""
    hook = paths.hooks.join(name)
    hook.write(
        '#!/bin/sh\n'
        f'echo HOOK:{name}\n'
        f'exit {code}\n'
    )
    hook.chmod(0o755)



"""Test init"""

import pytest


@pytest.mark.parametrize(
    'alt_work, repo_present, force', [
        (False, False, False),
        (True, False, False),
        (False, True, False),
        (False, True, True),
        (True, True, True),
    ], ids=[
        'simple',
        '-w',
        'existing repo',
        '-f',
        '-w & -f',
    ])
@pytest.mark.usefixtures('ds1_work_copy')
def test_init(
        runner, yadm_y, paths, repo_config, alt_work, repo_present, force):
    """Test init

    Repos should have attribs:
        - 0600 permissions
        - not bare
        - worktree = $HOME
        - showUntrackedFiles = no
        - yadm.managed = true
    """

    # these tests will assume this for $HOME
    home = str(paths.root.mkdir('HOME'))

    # ds1_work_copy comes WITH an empty repo dir present.
    old_repo = paths.repo.join('old_repo')
    if repo_present:
        # Let's put some data in it, so we can confirm that data is gone when
        # forced to be overwritten.
        old_repo.write('old repo data')
        assert old_repo.isfile()
    else:
        paths.repo.remove()

    # command args
    args = ['init']
    if alt_work:
        args.extend(['-w', paths.work])
    if force:
        args.append('-f')

    # run init
    run = runner(yadm_y(*args), env={'HOME': home})
    assert run.err == ''

    if repo_present and not force:
        assert run.failure
        assert 'repo already exists' in run.out
        assert old_repo.isfile(), 'Missing original repo'
    else:
        assert run.success
        assert 'Initialized empty shared Git repository' in run.out

        if repo_present:
            assert not old_repo.isfile(), 'Original repo still exists'

        if alt_work:
            assert repo_config('core.worktree') == paths.work
        else:
            assert repo_config('core.worktree') == home

        # uniform repo assertions
        assert oct(paths.repo.stat().mode).endswith('00'), (
            'Repo is not secure')
        assert repo_config('core.bare') == 'false'
        assert repo_config('status.showUntrackedFiles') == 'no'
        assert repo_config('yadm.managed') == 'true'



"""Test asserting private directories"""

import os
import re
import pytest

pytestmark = pytest.mark.usefixtures('ds1_copy')
PRIVATE_DIRS = ['.gnupg', '.ssh']


@pytest.mark.parametrize('home', [True, False], ids=['home', 'not-home'])
def test_pdirs_missing(runner, yadm_y, paths, home):
    """Private dirs (private dirs missing)

    When a git command is run
    And private directories are missing
    Create private directories prior to command
    """

    # confirm directories are missing at start
    for pdir in PRIVATE_DIRS:
        path = paths.work.join(pdir)
        if path.exists():
            path.remove()
        assert not path.exists()

    env = {'DEBUG': 'yes'}
    if home:
        env['HOME'] = paths.work

    # run status
    run = runner(command=yadm_y('status'), env=env)
    assert run.success
    assert run.err == ''
    assert 'On branch master' in run.out

    # confirm directories are created
    # and are protected
    for pdir in PRIVATE_DIRS:
        path = paths.work.join(pdir)
        if home:
            assert path.exists()
            assert oct(path.stat().mode).endswith('00'), ('Directory is '
                                                          'not secured')
        else:
            assert not path.exists()

    # confirm directories are created before command is run:
    if home:
        assert re.search(
            (r'Creating.+\.(gnupg|ssh).+Creating.+\.(gnupg|ssh).+'
             r'Running git command git status'),
            run.out, re.DOTALL), 'directories created before command is run'


def test_pdirs_missing_apd_false(runner, yadm_y, paths):
    """Private dirs (private dirs missing / yadm.auto-private-dirs=false)

    When a git command is run
    And private directories are missing
    But auto-private-dirs is false
    Do not create private dirs
    """

    # confirm directories are missing at start
    for pdir in PRIVATE_DIRS:
        path = paths.work.join(pdir)
        if path.exists():
            path.remove()
        assert not path.exists()

    # set configuration
    os.system(' '.join(yadm_y(
        'config', '--bool', 'yadm.auto-private-dirs', 'false')))

    # run status
    run = runner(command=yadm_y('status'))
    assert run.success
    assert run.err == ''
    assert 'On branch master' in run.out

    # confirm directories are STILL missing
    for pdir in PRIVATE_DIRS:
        assert not paths.work.join(pdir).exists()


def test_pdirs_exist_apd_false(runner, yadm_y, paths):
    """Private dirs (private dirs exist / yadm.auto-perms=false)

    When a git command is run
    And private directories exist
    And yadm is configured not to auto update perms
    Do not alter directories
    """

    # create permissive directories
    for pdir in PRIVATE_DIRS:
        path = paths.work.join(pdir)
        if not path.isdir():
            path.mkdir()
        path.chmod(0o777)
        assert oct(path.stat().mode).endswith('77'), 'Directory is secure.'

    # set configuration
    os.system(' '.join(yadm_y(
        'config', '--bool', 'yadm.auto-perms', 'false')))

    # run status
    run = runner(command=yadm_y('status'))
    assert run.success
    assert run.err == ''
    assert 'On branch master' in run.out

    # created directories are STILL permissive
    for pdir in PRIVATE_DIRS:
        path = paths.work.join(pdir)
        assert oct(path.stat().mode).endswith('77'), 'Directory is secure'



"""Unit tests: choose_template_cmd"""
import pytest


@pytest.mark.parametrize('label', ['', 'default', 'other'])
@pytest.mark.parametrize('awk', [True, False], ids=['awk', 'no-awk'])
def test_kind_default(runner, yadm, awk, label):
    """Test kind: default"""

    expected = 'template_default'
    awk_avail = 'true'

    if not awk:
        awk_avail = 'false'
        expected = ''

    if label == 'other':
        expected = ''

    script = f"""
        YADM_TEST=1 source {yadm}
        function awk_available {{ { awk_avail}; }}
        template="$(choose_template_cmd "{label}")"
        echo "TEMPLATE:$template"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert f'TEMPLATE:{expected}\n' in run.out


@pytest.mark.parametrize('label', ['envtpl', 'j2cli', 'j2', 'other'])
@pytest.mark.parametrize('envtpl', [True, False], ids=['envtpl', 'no-envtpl'])
@pytest.mark.parametrize('j2cli', [True, False], ids=['j2cli', 'no-j2cli'])
def test_kind_j2cli_envtpl(runner, yadm, envtpl, j2cli, label):
    """Test kind: j2 (both j2cli & envtpl)

    j2cli is preferred over envtpl if available.
    """

    envtpl_avail = 'true' if envtpl else 'false'
    j2cli_avail = 'true' if j2cli else 'false'

    if label in ('j2cli', 'j2') and j2cli:
        expected = 'template_j2cli'
    elif label in ('envtpl', 'j2') and envtpl:
        expected = 'template_envtpl'
    else:
        expected = ''

    script = f"""
        YADM_TEST=1 source {yadm}
        function envtpl_available {{ { envtpl_avail}; }}
        function j2cli_available {{ { j2cli_avail}; }}
        template="$(choose_template_cmd "{label}")"
        echo "TEMPLATE:$template"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    assert f'TEMPLATE:{expected}\n' in run.out



"""Unit tests: issue_legacy_path_warning"""
import pytest


@pytest.mark.parametrize(
    'legacy_path', [
        None,
        'repo.git',
        'config',
        'encrypt',
        'files.gpg',
        'bootstrap',
        'hooks/pre_command',
        'hooks/post_command',
        ],
    )
@pytest.mark.parametrize(
    'upgrade', [True, False], ids=['upgrade', 'no-upgrade'])
def test_legacy_warning(tmpdir, runner, yadm, upgrade, legacy_path):
    """Use issue_legacy_path_warning"""
    home = tmpdir.mkdir('home')

    if legacy_path:
        home.mkdir(f'.yadm').ensure(legacy_path)

    main_args = 'MAIN_ARGS=("upgrade")' if upgrade else ''
    script = f"""
        HOME={home}
        YADM_TEST=1 source {yadm}
        {main_args}
        issue_legacy_path_warning
        echo "LWI:$LEGACY_WARNING_ISSUED"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if legacy_path and not upgrade:
        assert 'Legacy configuration paths have been detected' in run.out
        assert 'LWI:1' in run.out
    else:
        assert run.out.rstrip() == 'LWI:0'



"""Test git"""

import re
import pytest


@pytest.mark.usefixtures('ds1_copy')
def test_git(runner, yadm_y, paths):
    """Test series of passthrough git commands

    Passthru unknown commands to Git
    Git command 'add' - badfile
    Git command 'add'
    Git command 'status'
    Git command 'commit'
    Git command 'log'
    """

    # passthru unknown commands to Git
    run = runner(command=yadm_y('bogus'))
    assert run.failure
    assert "git: 'bogus' is not a git command." in run.err
    assert "See 'git --help'" in run.err
    assert run.out == ''

    # git command 'add' - badfile
    run = runner(command=yadm_y('add', '-v', 'does_not_exist'))
    assert run.code == 128
    assert "pathspec 'does_not_exist' did not match any files" in run.err
    assert run.out == ''

    # git command 'add'
    newfile = paths.work.join('test_git')
    newfile.write('test_git')
    run = runner(command=yadm_y('add', '-v', str(newfile)))
    assert run.success
    assert run.err == ''
    assert "add 'test_git'" in run.out

    # git command 'status'
    run = runner(command=yadm_y('status'))
    assert run.success
    assert run.err == ''
    assert re.search(r'new file:\s+test_git', run.out)

    # git command 'commit'
    run = runner(command=yadm_y('commit', '-m', 'Add test_git'))
    assert run.success
    assert run.err == ''
    assert '1 file changed' in run.out
    assert '1 insertion' in run.out
    assert re.search(r'create mode .+ test_git', run.out)

    # git command 'log'
    run = runner(command=yadm_y('log', '--oneline'))
    assert run.success
    assert run.err == ''
    assert 'Add test_git' in run.out



"""Global tests configuration and fixtures"""

import collections
import contextlib
import copy
import distutils.dir_util  # pylint: disable=no-name-in-module,import-error
import os
import platform
import pwd
from subprocess import Popen, PIPE
import py
import pytest


def pytest_addoption(parser):
    """Add options to pytest"""
    parser.addoption(
        "--force-linters",
        action="store_true",
        default=False,
        help="Run linters regardless of installed versions",
    )


@pytest.fixture(scope='session')
def shellcheck_version():
    """Version of shellcheck supported"""
    return '0.4.6'


@pytest.fixture(scope='session')
def pylint_version():
    """Version of pylint supported"""
    return '2.4.1'


@pytest.fixture(scope='session')
def flake8_version():
    """Version of flake8 supported"""
    return '3.7.8'


@pytest.fixture(scope='session')
def yamllint_version():
    """Version of yamllint supported"""
    return '1.17.0'


@pytest.fixture(scope='session')
def tst_user():
    """Test session's user id"""
    return pwd.getpwuid(os.getuid()).pw_name


@pytest.fixture(scope='session')
def tst_host():
    """Test session's short hostname value"""
    return platform.node().split('.')[0]


@pytest.fixture(scope='session')
def tst_distro(runner):
    """Test session's distro"""
    distro = ''
    with contextlib.suppress(Exception):
        run = runner(command=['lsb_release', '-si'], report=False)
        distro = run.out.strip()
    return distro


@pytest.fixture(scope='session')
def tst_sys():
    """Test session's uname value"""
    return platform.system()


@pytest.fixture(scope='session')
def supported_commands():
    """List of supported commands

    This list should be updated every time yadm learns a new command.
    """
    return [
        'alt',
        'bootstrap',
        'clean',
        'clone',
        'config',
        'decrypt',
        'encrypt',
        'enter',
        'git-crypt',
        'gitconfig',
        'help',
        'init',
        'introspect',
        'list',
        'perms',
        'upgrade',
        'version',
        ]


@pytest.fixture(scope='session')
def supported_configs():
    """List of supported config options

    This list should be updated every time yadm learns a new config.
    """
    return [
        'local.class',
        'local.hostname',
        'local.os',
        'local.user',
        'yadm.alt-copy',
        'yadm.auto-alt',
        'yadm.auto-exclude',
        'yadm.auto-perms',
        'yadm.auto-private-dirs',
        'yadm.git-program',
        'yadm.gpg-perms',
        'yadm.gpg-program',
        'yadm.gpg-recipient',
        'yadm.ssh-perms',
        ]


@pytest.fixture(scope='session')
def supported_switches():
    """List of supported switches

    This list should be updated every time yadm learns a new switch.
    """
    return [
        '--yadm-archive',
        '--yadm-bootstrap',
        '--yadm-config',
        '--yadm-dir',
        '--yadm-encrypt',
        '--yadm-repo',
        '-Y',
        ]


@pytest.fixture(scope='session')
def supported_local_configs(supported_configs):
    """List of supported local config options"""
    return [c for c in supported_configs if c.startswith('local.')]


class Runner():
    """Class for running commands

    Within yadm tests, this object should be used when running commands that
    require:

      * Acting on the status code
      * Parsing the output of the command
      * Passing input to the command

    Other instances of simply running commands should use os.system().
    """

    def __init__(
            self,
            command,
            inp=None,
            shell=False,
            cwd=None,
            env=None,
            expect=None,
            report=True):
        if shell:
            self.command = ' '.join([str(cmd) for cmd in command])
        else:
            self.command = command
        self.inp = inp
        self.wrap(expect)
        process = Popen(
            self.command,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            shell=shell,
            cwd=cwd,
            env=env,
        )
        input_bytes = self.inp
        if self.inp:
            input_bytes = self.inp.encode()
        (out_bstream, err_bstream) = process.communicate(input=input_bytes)
        self.out = out_bstream.decode()
        self.err = err_bstream.decode()
        self.code = process.wait()
        self.success = self.code == 0
        self.failure = self.code != 0
        if report:
            self.report()

    def __repr__(self):
        return f'Runner({self.command})'

    def report(self):
        """Print code/stdout/stderr"""
        print(f'{self}')
        print(f'  RUN: code:{self.code}')
        if self.inp:
            print(f'  RUN: input:\n{self.inp}')
        print(f'  RUN: stdout:\n{self.out}')
        print(f'  RUN: stderr:\n{self.err}')

    def wrap(self, expect):
        """Wrap command with expect"""
        if not expect:
            return
        cmdline = ' '.join([f'"{w}"' for w in self.command])
        expect_script = f'set timeout 2\nspawn {cmdline}\n'
        for question, answer in expect:
            expect_script += (
                'expect {\n'
                f'"{question}" {{send "{answer}\\r"}}\n'
                'timeout {close;exit 128}\n'
                '}\n')
        expect_script += (
            'expect eof\n'
            'foreach {pid spawnid os_error_flag value} [wait] break\n'
            'exit $value')
        self.inp = expect_script
        print(f'EXPECT:{expect_script}')
        self.command = ['expect']


@pytest.fixture(scope='session')
def runner():
    """Class for running commands"""
    return Runner


@pytest.fixture(scope='session')
def config_git():
    """Configure global git configuration, if missing"""
    os.system(
        'git config user.name || '
        'git config --global user.name "test"')
    os.system(
        'git config user.email || '
        'git config --global user.email "test@test.test"')


@pytest.fixture()
def repo_config(runner, paths):
    """Function to query a yadm repo configuration value"""

    def query_func(key):
        """Query a yadm repo configuration value"""
        run = runner(
            command=('git', 'config', '--local', key),
            env={'GIT_DIR': paths.repo},
            report=False,
            )
        return run.out.rstrip()

    return query_func


@pytest.fixture(scope='session')
def yadm():
    """Path to yadm program to be tested"""
    full_path = os.path.realpath('yadm')
    assert os.path.isfile(full_path), "yadm program file isn't present"
    return full_path


@pytest.fixture()
def paths(tmpdir, yadm):
    """Function scoped test paths"""
    dir_root = tmpdir.mkdir('root')
    dir_work = dir_root.mkdir('work')
    dir_yadm = dir_root.mkdir('yadm')
    dir_repo = dir_yadm.mkdir('repo.git')
    dir_hooks = dir_yadm.mkdir('hooks')
    dir_remote = dir_root.mkdir('remote')
    file_archive = dir_yadm.join('files.gpg')
    file_bootstrap = dir_yadm.join('bootstrap')
    file_config = dir_yadm.join('config')
    file_encrypt = dir_yadm.join('encrypt')
    paths = collections.namedtuple(
        'Paths', [
            'pgm',
            'root',
            'work',
            'yadm',
            'repo',
            'hooks',
            'remote',
            'archive',
            'bootstrap',
            'config',
            'encrypt',
            ])
    return paths(
        yadm,
        dir_root,
        dir_work,
        dir_yadm,
        dir_repo,
        dir_hooks,
        dir_remote,
        file_archive,
        file_bootstrap,
        file_config,
        file_encrypt,
        )


@pytest.fixture()
def yadm_y(paths):
    """Generate custom command_list function"""
    def command_list(*args):
        """Produce params for running yadm with -Y"""
        return [paths.pgm, '-Y', str(paths.yadm)] + list(args)
    return command_list


class DataFile():
    """Datafile object"""

    def __init__(self, path, tracked=True, private=False):
        self.__path = path
        self.__parent = None
        self.__tracked = tracked
        self.__private = private

    @property
    def path(self):
        """Path property"""
        return self.__path

    @property
    def relative(self):
        """Relative path property"""
        if self.__parent:
            return self.__parent.join(self.path)
        raise BaseException('Unable to provide relative path, no parent')

    @property
    def tracked(self):
        """Tracked property"""
        return self.__tracked

    @property
    def private(self):
        """Private property"""
        return self.__private

    def relative_to(self, parent):
        """Update all relative paths to this py.path"""
        self.__parent = parent


class DataSet():
    """Dataset object"""

    def __init__(self):
        self.__files = list()
        self.__dirs = list()
        self.__tracked_dirs = list()
        self.__private_dirs = list()
        self.__relpath = None

    def __repr__(self):
        return (
            f'[DS with {len(self)} files; '
            f'{len(self.tracked)} tracked, '
            f'{len(self.private)} private]'
            )

    def __iter__(self):
        return iter(self.__files)

    def __len__(self):
        return len(self.__files)

    def __contains__(self, datafile):
        if [f for f in self.__files if f.path == datafile]:
            return True
        if datafile in self.__files:
            return True
        return False

    @property
    def files(self):
        """List of DataFiles in DataSet"""
        return list(self.__files)

    @property
    def tracked(self):
        """List of tracked DataFiles in DataSet"""
        return [f for f in self.__files if f.tracked]

    @property
    def private(self):
        """List of private DataFiles in DataSet"""
        return [f for f in self.__files if f.private]

    @property
    def dirs(self):
        """List of directories in DataSet"""
        return list(self.__dirs)

    @property
    def plain_dirs(self):
        """List of directories in DataSet not starting with '.'"""
        return [d for d in self.dirs if not d.startswith('.')]

    @property
    def hidden_dirs(self):
        """List of directories in DataSet starting with '.'"""
        return [d for d in self.dirs if d.startswith('.')]

    @property
    def tracked_dirs(self):
        """List of directories in DataSet not starting with '.'"""
        return [d for d in self.__tracked_dirs if not d.startswith('.')]

    @property
    def private_dirs(self):
        """List of directories in DataSet considered 'private'"""
        return list(self.__private_dirs)

    def add_file(self, path, tracked=True, private=False):
        """Add file to data set"""
        if path not in self:
            datafile = DataFile(path, tracked, private)
            if self.__relpath:
                datafile.relative_to(self.__relpath)
            self.__files.append(datafile)

        dname = os.path.dirname(path)
        if dname and dname not in self.__dirs:
            self.__dirs.append(dname)
            if tracked:
                self.__tracked_dirs.append(dname)
            if private:
                self.__private_dirs.append(dname)

    def relative_to(self, relpath):
        """Update all relative paths to this py.path"""
        self.__relpath = relpath
        for datafile in self.files:
            datafile.relative_to(self.__relpath)


@pytest.fixture(scope='session')
def ds1_dset(tst_sys):
    """Meta-data for dataset one files"""
    dset = DataSet()
    dset.add_file('t1')
    dset.add_file('d1/t2')
    dset.add_file(f'test_alt_copy##os.{tst_sys}')
    dset.add_file('u1', tracked=False)
    dset.add_file('d2/u2', tracked=False)
    dset.add_file('.ssh/p1', tracked=False, private=True)
    dset.add_file('.ssh/.p2', tracked=False, private=True)
    dset.add_file('.gnupg/p3', tracked=False, private=True)
    dset.add_file('.gnupg/.p4', tracked=False, private=True)
    return dset


@pytest.fixture(scope='session')
def ds1_data(tmpdir_factory, config_git, ds1_dset, runner):
    """A set of test data, worktree & repo"""
    # pylint: disable=unused-argument
    # This is ignored because
    # @pytest.mark.usefixtures('config_git')
    # cannot be applied to another fixture.

    data = tmpdir_factory.mktemp('ds1')

    work = data.mkdir('work')
    for datafile in ds1_dset:
        work.join(datafile.path).write(datafile.path, ensure=True)

    repo = data.mkdir('repo.git')
    env = os.environ.copy()
    env['GIT_DIR'] = str(repo)
    runner(
        command=['git', 'init', '--shared=0600', '--bare', str(repo)],
        report=False)
    runner(
        command=['git', 'config', 'core.bare', 'false'],
        env=env,
        report=False)
    runner(
        command=['git', 'config', 'status.showUntrackedFiles', 'no'],
        env=env,
        report=False)
    runner(
        command=['git', 'config', 'yadm.managed', 'true'],
        env=env,
        report=False)
    runner(
        command=['git', 'config', 'core.worktree', str(work)],
        env=env,
        report=False)
    runner(
        command=['git', 'add'] +
        [str(work.join(f.path)) for f in ds1_dset if f.tracked],
        env=env)
    runner(
        command=['git', 'commit', '--allow-empty', '-m', 'Initial commit'],
        env=env,
        report=False)

    data = collections.namedtuple('Data', ['work', 'repo'])
    return data(work, repo)


@pytest.fixture()
def ds1_work_copy(ds1_data, paths):
    """Function scoped copy of ds1_data.work"""
    distutils.dir_util.copy_tree(  # pylint: disable=no-member
        str(ds1_data.work), str(paths.work))


@pytest.fixture()
def ds1_repo_copy(runner, ds1_data, paths):
    """Function scoped copy of ds1_data.repo"""
    distutils.dir_util.copy_tree(  # pylint: disable=no-member
        str(ds1_data.repo), str(paths.repo))
    env = os.environ.copy()
    env['GIT_DIR'] = str(paths.repo)
    runner(
        command=['git', 'config', 'core.worktree', str(paths.work)],
        env=env,
        report=False)


@pytest.fixture()
def ds1_copy(ds1_work_copy, ds1_repo_copy):
    """Function scoped copy of ds1_data"""
    # pylint: disable=unused-argument
    # This is ignored because
    # @pytest.mark.usefixtures('ds1_work_copy', 'ds1_repo_copy')
    # cannot be applied to another fixture.
    return None


@pytest.fixture()
def ds1(ds1_work_copy, paths, ds1_dset):
    """Function scoped ds1_dset w/paths"""
    # pylint: disable=unused-argument
    # This is ignored because
    # @pytest.mark.usefixtures('ds1_copy')
    # cannot be applied to another fixture.
    dscopy = copy.deepcopy(ds1_dset)
    dscopy.relative_to(copy.deepcopy(paths.work))
    return dscopy


@pytest.fixture(scope='session')
def gnupg(tmpdir_factory, runner):
    """Location of GNUPGHOME"""

    def register_gpg_password(password):
        """Publish a new GPG mock password"""
        py.path.local('/tmp/mock-password').write(password)

    home = tmpdir_factory.mktemp('gnupghome')
    home.chmod(0o700)
    conf = home.join('gpg.conf')
    conf.write('no-secmem-warning\n')
    conf.chmod(0o600)
    agentconf = home.join('gpg-agent.conf')
    agentconf.write(
        f'pinentry-program {os.path.abspath("test/pinentry-mock")}\n'
        'max-cache-ttl 0\n'
    )
    agentconf.chmod(0o600)
    data = collections.namedtuple('GNUPG', ['home', 'pw'])
    env = os.environ.copy()
    env['GNUPGHOME'] = home

    # this pre-populates std files in the GNUPGHOME
    runner(['gpg', '-k'], env=env)

    return data(home, register_gpg_password)



"""Unit tests: set_yadm_dir"""
import pytest


@pytest.mark.parametrize(
    'condition',
    ['basic', 'override', 'xdg_config_home', 'legacy'],
    )
def test_set_yadm_dir(runner, yadm, condition):
    """Test set_yadm_dir"""
    setup = ''
    if condition == 'override':
        setup = 'YADM_DIR=/override'
    elif condition == 'xdg_config_home':
        setup = 'XDG_CONFIG_HOME=/xdg'
    elif condition == 'legacy':
        setup = 'YADM_COMPATIBILITY=1'
    script = f"""
        HOME=/testhome
        YADM_TEST=1 source {yadm}
        {setup}
        set_yadm_dir
        echo "$YADM_DIR"
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if condition == 'basic':
        assert run.out.rstrip() == '/testhome/.config/yadm'
    elif condition == 'override':
        assert run.out.rstrip() == '/override'
    elif condition == 'xdg_config_home':
        assert run.out.rstrip() == '/xdg/yadm'
    elif condition == 'legacy':
        assert run.out.rstrip() == '/testhome/.yadm'



"""Unit tests: set_operating_system"""

import pytest


@pytest.mark.parametrize(
    'proc_value, expected_os', [
        ('missing', 'uname'),
        ('has Microsoft inside', 'WSL'),
        ('another value', 'uname'),
    ], ids=[
        '/proc/version missing',
        '/proc/version includes MS',
        '/proc/version excludes MS',
    ])
def test_set_operating_system(
        runner, paths, tst_sys, proc_value, expected_os):
    """Run set_operating_system and test result"""

    # Normally /proc/version (set in PROC_VERSION) is inspected to identify
    # WSL. During testing, we will override that value.
    proc_version = paths.root.join('proc_version')
    if proc_value != 'missing':
        proc_version.write(proc_value)
    script = f"""
        YADM_TEST=1 source {paths.pgm}
        PROC_VERSION={proc_version}
        set_operating_system
        echo $OPERATING_SYSTEM
    """
    run = runner(command=['bash'], inp=script)
    assert run.success
    assert run.err == ''
    if expected_os == 'uname':
        expected_os = tst_sys
    assert run.out.rstrip() == expected_os
